#!/usr/bin/env python

from optparse import OptionParser
import socket
import subprocess
import sys
import os
try:
    from .waterfall import Waterfall
except:
    from blimpy import Waterfall

#import pdb #pdb.set_trace()

local_host = socket.gethostname()

def reset_outs():
    """ Returns None, None, used as a helper """

    return None,None

def make_batch_script():
    """ This creates a batch script for the heavy lifting"""

    script_text  = '#! /bin/bash\n# This script calculates the md5sum for the last number of bits (nbits).# Usage :\n#     ./tail_sum file bnits\n#\n\nfile_name=$1\nbit_num=$2\n\ntail -c $bit_num $file_name | md5sum\n'

    with open('tail_sum.sh', 'w') as batch_script:
        batch_script.write(script_text)

    os.chmod('tail_sum.sh', 0o775)

def find_header_size(filename):
    """Script to find the header size of a filterbank file

    Args:
        filename (str): Name of file

    Returns size of header in Bytes (int)
    """

    # open datafile
    filfile=open(filename,'rb')
    # go to the start of the file
    filfile.seek(0)
    #read some region larger than the header.
    round1 = filfile.read(1000)
    headersize = round1.find(b'HEADER_END')+len(b'HEADER_END')

    return headersize

def cmd_tool(args=None):
    """ Command line tool to make a MD5sum comparison of two .fil files.

    Supplies a command line:
        matchfils <FIL_FILE1> <FIL_FILE2>
    which will compare the two files to confirm the files have matching MD5sums

    Notes:
        This command is specific to BL @ Green Bank
    """

    if 'bl' in local_host:
        header_loc = '/usr/local/sigproc/bin/header' #Current location of header command in GBT.
    else:
        raise IOError('Script only able to run in BL systems.')

    p = OptionParser()
    p.set_usage('matchfils <FIL_FILE1> <FIL_FILE2>')
    opts, args = p.parse_args(sys.argv[1:])

    file1 = args[0]
    file2 = args[1]

    #------------------------------------
    #Create batch script

    make_batch_script()

    #------------------------------------
    #First checksum

    headersize1 = find_header_size(file1)
    file_size1 = os.path.getsize(file1)

    #Strip header from file, and calculate the md5sum of the rest.
    #command=['tail','-c',str(file_size1-headersize1),file1,'|','md5sum']
    command=['./tail_sum.sh',file1,str(file_size1-headersize1)]
    print('[matchfils] '+' '.join(command))

    proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out, err) = proc.communicate()

    check_sum1 = out.split()[0]
    print('[matchfils] Checksum is:', check_sum1)

    if err:
        raise Error('There is an error.')

    #---
    out,err = reset_outs()

    command=[header_loc,file1]
    print('[matchfils] Header information:')

    proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out, err) = proc.communicate()

    header1 = out
    print(header1)

    #------------------------------------
    #Second checksum

    out,err = reset_outs()

    headersize2 = find_header_size(file2)
    file_size2 = os.path.getsize(file2)

    #Strip header from file, and calculate the md5sum of the rest.
    command=['./tail_sum.sh',file2,str(file_size2-headersize2)]
    print('[matchfils] '+' '.join(command))

    proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out, err) = proc.communicate()

    check_sum2 = out.split()[0]
    print('[matchfils] Checksum is:', check_sum2)

    if err:
        raise Error('There is an error.')

    #---
    out,err = reset_outs()

    command=[header_loc,file2]
    print('[matchfils] Header information:')

    proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out, err) = proc.communicate()

    header2 = out
    print(header2)

    #------------------------------------
    #check the checksums

    if check_sum1 != check_sum2:
        print('[matchfils] Booo! Checksum does not match between files.')
    else:
        print('[matchfils] Hooray! Checksum matches between files.')

    #------------------------------------
    #Remove batch script

    os.remove('tail_sum.sh')

if __name__ == "__main__":
    cmd_tool()




from .config import *
from .compute_lst import compute_lst

def compute_lsrk(wf):
    """ Computes the LSR in km/s

    Computes the Local standard of rest kinematic using the time (MJD),
    RA and DEC of the observation to compute along with the telescope location.
    Requires pyslalib

    Args:
        wf (bl.Waterfall): Waterfall object for which to compute LSR
    """
    ra = Angle(wf.header['src_raj'], unit='hourangle')
    dec = Angle(wf.header['src_dej'], unit='degree')
    mjdd = wf.header['tstart']
    rarad = ra.to('radian').value
    dcrad = dec.to('radian').value
    last = compute_lst(wf)
    tellat = np.deg2rad(wf.coords[0])
    tellong = np.deg2rad(wf.coords[1])

    # convert star position to vector
    starvect = s.sla_dcs2c(rarad, dcrad)

    # velocity component in ra,dec due to Earth rotation
    Rgeo = s.sla_rverot(tellat, rarad, dcrad, last)

    # get Barycentric and heliocentric velocity and position of the Earth.
    evp = s.sla_evp(mjdd, 2000.0)
    dvb = evp[0]  # barycentric velocity vector, in AU/sec
    dpb = evp[1]  # barycentric position vector, in AU
    dvh = evp[2]  # heliocentric velocity vector, in AU/sec
    dph = evp[3]  # heliocentric position vector, in AU

    # dot product of vector to object and heliocentric velocity
    # convert AU/sec to km/sec
    vcorhelio = -s.sla_dvdv(starvect, dvh) * 149.597870e6
    vcorbary = -s.sla_dvdv(starvect, dvb) * 149.597870e6

    # rvlsrd is velocity component in ra,dec direction due to the Sun's
    # motion with respect to the "dynamical" local standard of rest
    rvlsrd = s.sla_rvlsrd(rarad, dcrad)

    # rvlsrk is velocity component in ra,dec direction due to i
    # the Sun's motion w.r.t the "kinematic" local standard of rest
    rvlsrk = s.sla_rvlsrk(rarad, dcrad)

    # rvgalc is velocity component in ra,dec direction due to
    # the rotation of the Galaxy.
    rvgalc = s.sla_rvgalc(rarad, dcrad)
    totalhelio = Rgeo + vcorhelio
    totalbary = Rgeo + vcorbary
    totallsrk = totalhelio + rvlsrk
    totalgal = totalbary + rvlsrd + rvgalc

    return totallsrk



from .config import *
from ..utils import rebin, db
from .plot_utils import calc_extent

def plot_time_series(wf, f_start=None, f_stop=None, if_id=0, logged=True, orientation='h', MJD_time=False, **kwargs):
    """ Plot the time series.

     Args:
        f_start (float): start frequency, in MHz
        f_stop (float): stop frequency, in MHz
        logged (bool): Plot in linear (False) or dB units (True),
        kwargs: keyword args to be passed to matplotlib imshow()
    """

    ax = plt.gca()
    plot_f, plot_data = wf.grab_data(f_start, f_stop, if_id)

    # Since the data has been squeezed, the axis for time goes away if only one bin, causing a bug with axis=1
    if len(plot_data.shape) > 1:
        plot_data = np.nanmean(plot_data, axis=1)
    else:
        plot_data = np.nanmean(plot_data)

    if logged and wf.header['nbits'] >= 8:
        plot_data = db(plot_data)

    # Make proper time axis for plotting (but only for plotting!). Note that this makes the values inclusive.
    extent = calc_extent(wf, plot_f=plot_f, plot_t=wf.timestamps, MJD_time=MJD_time)
    plot_t = np.linspace(extent[2], extent[3], len(wf.timestamps))

    if MJD_time:
        tlabel = "Time [MJD]"
    else:
        tlabel = "Time [s]"

    if logged:
        plabel = "Power [dB]"
    else:
        plabel = "Power [counts]"

    # Reverse oder if vertical orientation.
    if 'v' in orientation:
        plt.plot(plot_data, plot_t, **kwargs)
        plt.xlabel(plabel)

    else:
        plt.plot(plot_t, plot_data, **kwargs)
        plt.xlabel(tlabel)
        plt.ylabel(plabel)

    ax.autoscale(axis='both', tight=True)



from __future__ import absolute_import


try:
    from . import fluxcal
    from . import stokescal
    from . import calib_plots
except ImportError:
    print("Warning: Cannot import calibration utilities")





import subprocess
import sys
from os import path, listdir

print("Reached blimpy.tests init!")
here = path.dirname(path.abspath(__file__))
print("Running tests from {}".format(here))
if "test_data" not in listdir(here):
    print("Test data has not yet been downloaded. Downloading Data...")
    if sys.version_info >= (3, 0):
        subprocess.run(["sh", "download_data.sh"])
    else:
        subprocess.call(["sh", "download_data.sh"])



"""
Test the calcload.py utility and function calc_max_load()
"""
from tests.data import voyager_h5, voyager_fil
from blimpy.calcload import cmd_tool, calc_max_load

def test_calcload():
    r""" Test the calcload command line tool """
    args = [voyager_h5]
    cmd_tool(args)
    args = ['-v', voyager_fil]
    cmd_tool(args)

def test_calc_max_load():
    gb1 = calc_max_load(voyager_h5)
    gb2 = calc_max_load(voyager_fil)
    assert(gb1 == gb2 == 1.0)
   
if __name__ == "__main__":
    test_calcload()
    test_calc_max_load()



import os
import numpy as np
from tests.data import voyager_h5, voyager_fil
import blimpy as bl

import pytest

OUTDIR = os.path.dirname(voyager_h5) + "/"

def test_info():
    a = bl.Waterfall(voyager_h5)
    print(a)
    a.info()
    a.blank_dc(n_coarse_chan=1)
    a.calibrate_band_pass_N1()
    # Below: orphaned functions (not used anywhere else).
    # That is a little strange...
    a.grab_data()
    a.read_data()
    # It is VERY strange for INTERNAL functions to be orphaned!
    a._get_chunk_dimensions()
    # plenty of missed if suites
    a._get_blob_dimensions((300, 300, 300, 300))
    a._update_header()

def test_get_freqs():
    wf = bl.Waterfall(voyager_h5)
    freqs = wf.container.populate_freqs()
    sum1 = np.sum(freqs)
    freqs = wf.get_freqs()
    sum2 = np.sum(freqs)
    assert sum1 == sum2
    wf = bl.Waterfall(voyager_fil)
    freqs = wf.container.populate_freqs()
    sum1 = np.sum(freqs)
    freqs = wf.get_freqs()
    sum2 = np.sum(freqs)
    assert sum1 == sum2

def test_cmdline():
    from blimpy.waterfall import cmd_tool

    args = [voyager_h5, '-S', '-s', OUTDIR + 'test.png']
    cmd_tool(args)

    args = [voyager_h5, '-i']
    cmd_tool(args)

    if os.path.exists(OUTDIR + 'test.h5'):
        os.remove(OUTDIR + 'test.h5')
        args = [voyager_h5, '-H', '-o', OUTDIR + 'test.h5']
        cmd_tool(args)
        assert os.path.exists(OUTDIR + 'test.h5')
        os.remove(OUTDIR + 'test.h5')

    if os.path.exists(OUTDIR + 'test.fil'):
        os.remove(OUTDIR + 'test.fil')
        args = [voyager_h5, '-F', '-o', OUTDIR + 'test.fil']
        cmd_tool(args)
        assert os.path.exists(OUTDIR + 'test.fil')
        os.remove(OUTDIR + 'test.fil')

def test_cmd_arguments():
    from blimpy.waterfall import cmd_tool
    
    args = [voyager_h5, '-H', '-F', '-o', OUTDIR + 'test.fil']
    with pytest.raises(ValueError):
        cmd_tool(args)

if __name__ == "__main__":
    test_info()
    test_cmdline()
    test_cmd_arguments()



from .config import *
import scipy.stats


def plot_kurtosis(wf, f_start=None, f_stop=None, if_id=0, **kwargs):
    """ Plot kurtosis

     Args:
        f_start (float): start frequency, in MHz
        f_stop (float): stop frequency, in MHz
        kwargs: keyword args to be passed to matplotlib imshow()
    """
    ax = plt.gca()

    plot_f, plot_data = wf.grab_data(f_start, f_stop, if_id)

    # Using accending frequency for all plots.
    if wf.header['foff'] < 0:
        plot_data = plot_data[..., ::-1]  # Reverse data
        plot_f = plot_f[::-1]

    try:
        pltdata = scipy.stats.kurtosis(plot_data, axis=0, nan_policy='omit')
    except:
        pltdata = plot_data * 0.0

    plt.plot(plot_f, pltdata, **kwargs)
    plt.ylabel("Kurtosis")
    plt.xlabel("Frequency [MHz]")

    plt.xlim(plot_f[0], plot_f[-1])



import h5py
import hdf5plugin
import numpy
import tempfile


def test_is_h5py_correctly_installed():
    """
    If this test fails you probably need to install h5py from source manually:

    $ pip install --no-binary=h5py h5py
    """
    f = h5py.File(tempfile.gettempdir() + '/h5testfile', "w")
    block_size = 0
    dataset = f.create_dataset(
        "data",
        (100, 100, 100),
        dtype='float32',
        **hdf5plugin.Bitshuffle(nelems=0, lz4=True)
    )

    array = numpy.random.rand(100, 100, 100)
    array = array.astype('float32')
    dataset[:] = array
    f.close()




from os.path import dirname
from blimpy.stix import cmd_tool
from tests.data import voyager_fil


PLOT_DIR = dirname(voyager_fil)


def execute_command(args):
    print("\ntest_stix: args:", args)
    cmd_tool(args)


def test_stix():

    args = [voyager_fil, "16", "--plot_dir",  PLOT_DIR]
    execute_command(args)

    args = [voyager_fil, "4", "--plot_dir",  PLOT_DIR, "-s", "v"]
    execute_command(args)

    args = [voyager_fil, "4", "-p",  PLOT_DIR, "--stitch", "h"]
    execute_command(args)

    args = [voyager_fil, "4", "--plot_dir",  PLOT_DIR, "--stitch", "n",
            "--dpi", "100", "--width", "8", "--height", "6"]
    execute_command(args)

    args = [voyager_fil, "4", "-p",  PLOT_DIR, "-s", "n", "-d", "100", "-w", "8", "-t", "6"]
    execute_command(args)


if __name__ == "__main__":
    test_stix()



from .config import *

def calc_extent(self, plot_f=None, plot_t=None, MJD_time=False):
    """ Setup plotting edges.
    """

    plot_f_begin = plot_f[0]
    plot_f_end = plot_f[-1] + (plot_f[1] - plot_f[0])

    plot_t_begin = self.timestamps[0]
    plot_t_end = self.timestamps[-1] + (self.timestamps[1] - self.timestamps[0])

    if MJD_time:
        extent = (plot_f_begin, plot_f_end, plot_t_begin, plot_t_end)
    else:
        extent = (plot_f_begin, plot_f_end, 0.0, (plot_t_end - plot_t_begin) * 24. * 60. * 60)

    return extent


import blimpy as bl
import numpy as np
from pprint import pprint
import pylab as plt

from tests.data import voyager_fil, voyager_h5

from blimpy.ephemeris import compute_lst, compute_lsrk

def test_compute_lst():
    """ Load Voyager dataset and test plotting """
    print("Loading HDF5 data with Waterfall()..."),
    a = bl.Waterfall(voyager_h5)
    print(compute_lst(a))


def test_compute_lsrk():
    a = bl.Waterfall(voyager_h5)
    print(compute_lsrk(a))


if __name__ == "__main__":
    test_compute_lst()
    test_compute_lsrk()


from .config import *
from . import plot_time_series, plot_kurtosis, plot_spectrum_min_max, plot_waterfall, plot_spectrum
from astropy import units as u


def plot_all(wf, t=0, f_start=None, f_stop=None, logged=False, if_id=0, kurtosis=True, **kwargs):
    """ Plot waterfall of data as well as spectrum; also, placeholder to make even more complicated plots in the future.

    Args:
        f_start (float): start frequency, in MHz
        f_stop (float): stop frequency, in MHz
        logged (bool): Plot in linear (False) or dB units (True),
        t (int): integration number to plot (0 -> len(data))
        logged (bool): Plot in linear (False) or dB units (True)
        if_id (int): IF identification (if multiple IF signals in file)
        kwargs: keyword args to be passed to matplotlib plot() and imshow()
    """
    if wf.header['nbits'] <= 2:
        logged = False

    nullfmt = NullFormatter()  # no labels

    # definitions for the axes
    left, width = 0.35, 0.5
    bottom, height = 0.45, 0.5
    width2, height2 = 0.1125, 0.15
    bottom2, left2 = bottom - height2 - .025, left - width2 - .02
    bottom3, left3 = bottom2 - height2 - .025, 0.075

    rect_waterfall = [left, bottom, width, height]
    rect_colorbar = [left + width, bottom, .025, height]
    rect_spectrum = [left, bottom2, width, height2]
    rect_min_max = [left, bottom3, width, height2]
    rect_timeseries = [left + width, bottom, width2, height]
    rect_kurtosis = [left3, bottom3, 0.25, height2]
    rect_header = [left3 - .05, bottom, 0.2, height]

    # --------
    #         axColorbar = plt.axes(rect_colorbar)
    #         print 'Ploting Colorbar'
    #         print plot_data.max()
    #         print plot_data.min()
    #
    #         plot_colorbar = range(plot_data.min(),plot_data.max(),int((plot_data.max()-plot_data.min())/plot_data.shape[0]))
    #         plot_colorbar = np.array([[plot_colorbar],[plot_colorbar]])
    #
    #         plt.imshow(plot_colorbar,aspect='auto', rasterized=True, interpolation='nearest',)

    #         axColorbar.xaxis.set_major_formatter(nullfmt)
    #         axColorbar.yaxis.set_major_formatter(nullfmt)

    #         heatmap = axColorbar.pcolor(plot_data, edgecolors = 'none', picker=True)
    #         plt.colorbar(heatmap, cax = axColorbar)
    # --------
    
    axMinMax = plt.axes(rect_min_max)
    print('Plotting Min Max')
    plot_spectrum_min_max(wf, logged=logged, f_start=f_start, f_stop=f_stop, t=t)
    plt.title('')
    axMinMax.yaxis.tick_right()
    axMinMax.yaxis.set_label_position("right")

    # --------
    axSpectrum = plt.axes(rect_spectrum,sharex=axMinMax)
    print('Plotting Spectrum')
    plot_spectrum(wf, logged=logged, f_start=f_start, f_stop=f_stop, t=t)
    plt.title('')
    axSpectrum.yaxis.tick_right()
    axSpectrum.yaxis.set_label_position("right")
    plt.xlabel('')
#        axSpectrum.xaxis.set_major_formatter(nullfmt)
    plt.setp(axSpectrum.get_xticklabels(), visible=False)

    # --------
    axWaterfall = plt.axes(rect_waterfall,sharex=axMinMax)
    print('Plotting Waterfall')
    plot_waterfall(wf, f_start=f_start, f_stop=f_stop, logged=logged, cb=False)
    plt.xlabel('')

    # no labels
#        axWaterfall.xaxis.set_major_formatter(nullfmt)
    plt.setp(axWaterfall.get_xticklabels(), visible=False)

    # --------
    axTimeseries = plt.axes(rect_timeseries)
    print('Plotting Timeseries')
    plot_time_series(wf, f_start=f_start, f_stop=f_stop, orientation='v')
    axTimeseries.yaxis.set_major_formatter(nullfmt)
#        axTimeseries.xaxis.set_major_formatter(nullfmt)

    # --------
    # Could exclude since it takes much longer to run than the other plots.
    if kurtosis:
        axKurtosis = plt.axes(rect_kurtosis)
        print('Plotting Kurtosis')
        plot_kurtosis(wf, f_start=f_start, f_stop=f_stop)


    # --------
    axHeader = plt.axes(rect_header)
    print('Plotting Header')
    # Generate nicer header
    telescopes = {0: 'Fake data',
                  1: 'Arecibo',
                  2: 'Ooty',
                  3: 'Nancay',
                  4: 'Parkes',
                  5: 'Jodrell',
                  6: 'GBT',
                  8: 'Effelsberg',
                  10: 'SRT',
                  64: 'MeerKAT',
                  65: 'KAT7'
                  }

    telescope = telescopes.get(wf.header['telescope_id'], wf.header['telescope_id'])

    plot_header = "%14s: %s\n" % ("TELESCOPE_ID", telescope)
    for key in ('SRC_RAJ', 'SRC_DEJ', 'TSTART', 'NCHANS', 'NBEAMS', 'NIFS', 'NBITS'):
        try:
            plot_header += "%14s: %s\n" % (key, wf.header[key.lower()])
        except KeyError:
            pass
    fch1 = "%6.6f MHz" % wf.header['fch1']

    foff = (wf.header['foff'] * 1e6 * u.Hz)
    if np.abs(foff) > 1e6 * u.Hz:
        foff = str(foff.to('MHz'))
    elif np.abs(foff) > 1e3 * u.Hz:
        foff = str(foff.to('kHz'))
    else:
        foff = str(foff.to('Hz'))

    plot_header += "%14s: %s\n" % ("FCH1", fch1)
    plot_header += "%14s: %s\n" % ("FOFF", foff)

    plt.text(0.05, .95, plot_header, ha='left', va='top', wrap=True)

    axHeader.set_facecolor('white')
    axHeader.xaxis.set_major_formatter(nullfmt)
    axHeader.yaxis.set_major_formatter(nullfmt)



import blimpy as bl
import numpy as np
from pprint import pprint
from tests.data import voyager_fil, voyager_h5
from blimpy.plotting.config import plt

def compare_waterfall_fil_to_h5():
    """ Load Voyager dataset and test that both fil and hdf5 readers return same headers and data """

    print("Loading FIL and HDF5 data with Waterfall()..."),
    a = bl.Waterfall(voyager_h5)
    b = bl.Waterfall(voyager_fil)
    print("OK")

    print("Reading headers..")
    print("\nHDF5 file header:")
    pprint(a.header)
    print("\nFIL file header:")
    pprint(b.header)
    print("Headers are loading OK")

    print("\nChecking header values match..."),
    for key in b.header.keys():
        assert b.header[key] == a.header[key]
    print("OK")

    print("Checking datatype matches..."),
    assert a.data.dtype == b.data.dtype
    print("OK")

    print("Checking data matches..."),
    assert np.allclose(a.data, b.data)
    assert a.data.dtype == b.data.dtype
    print("OK")

def compare_waterfall_fil_to_h5_methods_and_attributes():
    """ Compare attributes and check methods """
    a = bl.Waterfall(voyager_h5)
    b = bl.Waterfall(voyager_fil)

    print("Comparing attributes of classes match where expected")
    assert a.beam_axis == b.beam_axis
    assert a.freq_axis == b.freq_axis
    assert a.time_axis == b.time_axis
    #assert a.stokes_axis == b.stokes_axis  # Fil shouldn't have stokes axis ...

    assert a.calc_n_coarse_chan() == b.calc_n_coarse_chan()
    assert a.file_shape == b.file_shape

    assert a.n_channels_in_file == b.n_channels_in_file
    assert a.n_ints_in_file == b.n_ints_in_file
    assert a.selection_shape == b.selection_shape

    print("Checking if basic methods run without raising Exceptions")
    # Check they can be run
    a.container.populate_freqs()
    a.container.populate_timestamps()
    a.info()
    a.blank_dc(1)
    a.calibrate_band_pass_N1()

    b.container.populate_freqs()
    b.container.populate_timestamps()
    b.info()
    b.blank_dc(1)
    b.calibrate_band_pass_N1()

    # TODO: 'compute_lsrk' -- need PYSLALIB
    # TODO: 'compute_lst' -- need PYSLALIB

    # Expected to differ between: 'ext', 'file_size_bytes', 'filename', 'container'
    # Unused? Inherited from Filterbank.py 'gen_from_header', 'generate_freqs', 'read_filterbank', 'read_hdf5'

    # TODO: 'grab_data',
    # TODO: 'read_data',
    # TODO: 'write_to_fil',
    # TODO: 'write_to_filterbank',
    # TODO: 'write_to_hdf5']

    dir_a = dir(a)
    dir_b = dir(b)

    print("Attr/methods in HDF5 but not in FIL:")
    for item in dir_a:
        if item not in dir_b:
            print(item)

    print("Attr/methods in FIL but not in HDF5:")
    for item in dir_b:
        if item not in dir_a:
            print(item)

def compare_waterfall_fil_h5_conatiners():
    """ Compare the two containers for fil.container and h5.container """
    a = bl.Waterfall(voyager_h5)
    b = bl.Waterfall(voyager_fil)

    dir_a = dir(a.container)
    dir_b = dir(b.container)

    print("Attr/methods in HDF5 container but not in FIL:")
    for item in dir_a:
        if item not in dir_b:
            print(item)

    print("Attr/methods in FIL container but not in HDF5:")
    for item in dir_b:
        if item not in dir_a:
            print(item)

def test_plotting_doesnt_cause_exceptions():
    """ Try running the plotting routines. They should not raise expections even without X windows """
    a = bl.Waterfall(voyager_h5)
    b = bl.Waterfall(voyager_fil)

    a.plot_all()
    plt.clf()
    a.plot_kurtosis()
    plt.clf()
    a.plot_spectrum()
    plt.clf()
    a.plot_spectrum_min_max()
    plt.clf()
    a.plot_waterfall()
    plt.clf()
    a.plot_time_series()
    plt.clf()

    b.plot_all()
    plt.clf()
    b.plot_kurtosis()
    plt.clf()
    b.plot_spectrum()
    plt.clf()
    b.plot_spectrum_min_max()
    plt.clf()
    b.plot_waterfall()
    plt.clf()
    b.plot_time_series()
    plt.clf()

if __name__ == "__main__":
    compare_waterfall_fil_to_h5()
    compare_waterfall_fil_to_h5_methods_and_attributes()
    compare_waterfall_fil_h5_conatiners()
    test_plotting_doesnt_cause_exceptions()


#!/usr/bin/env python
"""
    Simple script for quicly making an h5 file from a .fil.

    ..author: Emilio Enriquez (jeenriquez@gmail.com)

    July 28th 2017
"""

import sys
import os
from argparse import ArgumentParser

# Logging set up
import logging
logger = logging.getLogger(__name__)

level_log = logging.INFO

if level_log == logging.INFO:
    stream = sys.stdout
    fmt = '%(name)-15s %(levelname)-8s %(message)s'
else:
    stream =  sys.stderr
    fmt = '%%(relativeCreated)5d (name)-15s %(levelname)-8s %(message)s'
logging.basicConfig(format=fmt,stream=stream,level = level_log)


from blimpy import Waterfall
from blimpy.io.hdf_writer import __write_to_hdf5_heavy as write_to_h5


def make_h5_file(filename, out_dir='./', new_filename=None, t_start=None, t_stop=None):
    """ Converts file to HDF5 (.h5) format. Default saves output in current dir.

    Args:
        filename (str): Name of filterbank file to read
        out_dir (str): Output directory path. Defaults to cwd
        new_filename (None or str): Name of output filename. If not set, will default
                                    to same as input, but with .h5 instead of .fil
        t_start (int): Start integration ID to be extracted from file
        t_stop (int): Stop integration ID to be extracted from file
    """

    wf = Waterfall(filename, load_data=False, t_start=t_start, t_stop=t_stop)
    if not new_filename:
        new_filename = out_dir+filename.replace('.fil', '.h5').split('/')[-1]

    if '.h5' not in new_filename:
        new_filename = new_filename+'.h5'

    write_to_h5(wf, new_filename)


def cmd_tool(args=None):
    """ Command line utility for converting Sigproc filterbank (.fil) to  HDF5 (.h5) format

    Usage:
        fil2h5 <FULL_PATH_TO_FIL_FILE> [options]

    Options:
      -h, --help            show this help message and exit
      -o OUT_DIR, --out_dir=OUT_DIR
                            Location for output files. Default: local dir.
      -n NEW_FILENAME, --new_filename=NEW_FILENAME
                            New name. Default: replaces extention to .h5
      -d, --delete_input    This option deletes the input file after conversion.
      -l MAX_LOAD           Maximum data limit to load. Default:1GB
    """

    parser = ArgumentParser(description="Command line utility for converting Sigproc filterbank (.fil) to  HDF5 (.h5) format  \n >>fil2h5 <FULL_PATH_TO_FIL_FILE> [options]")
    parser.add_argument("filepath_in", type=str, help="Path of input Filterbank file")
    parser.add_argument('-o', '--out_dir', dest='out_dir', type=str, default='./', help='Location for output files. Default: local dir. ')
    parser.add_argument('-n', '--new_filename', dest='new_filename', type=str, default='', help='New name. Default: replaces extention to .h5')
    parser.add_argument('-d', '--delete_input', dest='delete_input', action='store_true', default=False, help='This option deletes the input file after conversion.')
    parser.add_argument('-s', '--start_id', dest='t_start', type=int, default=None, help='start integration ID')
    parser.add_argument('-t', '--stop_id', dest='t_stop', type=int, default=None, help='stop integration ID')

    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    make_h5_file(args.filepath_in,
                 out_dir = args.out_dir,
                 new_filename = args.new_filename,
                 t_start=args.t_start,
                 t_stop=args.t_stop)

    if args.delete_input:
        logger.info("'Deleting input file: {}".format(args.filename_in))
        os.remove(args.filename_in)


if __name__ == "__main__":
    cmd_tool()



"""
# test_fil2h5
"""

import pytest

import os
import blimpy as bl
from tests.data import voyager_fil


OUTDIR = os.path.dirname(voyager_fil) + "/"


def test_fil2h5_conversion():
    """ Tests the conversion of fil files into h5 in both light and heavy modes.
    """

    # Creating test file.
    bl.fil2h5.make_h5_file(voyager_fil, new_filename='test.h5')

    # Testing filename
    bl.fil2h5.make_h5_file(voyager_fil, new_filename='test')

    # Deleting test file
    os.remove('test.h5')

def test_cmd_tool():
    """
    This is the same test file, but now through the cmd tool.
    """
    #with pytest.raises(SystemExit):
    args = [voyager_fil, '-n', OUTDIR + 'cmd.h5']
    bl.fil2h5.cmd_tool(args=args)

def test_no_args():
    """
    The cmd tool needs to exit, mandating a file name.
    """
    with pytest.raises(SystemExit):
        bl.fil2h5.cmd_tool("")


if __name__ == "__main__":
    test_fil2h5_conversion()
    test_cmd_tool()
    test_no_args()



r""" h5diag """

import os
import sys
from argparse import ArgumentParser
import h5py
from astropy.coordinates import Angle
from blimpy.io.hdf_reader import examine_h5


def oops(msg):
    print("\n*** h5diag: {}\n".format(msg))
    sys.exit(86)


def read_header(h5):
    """ Read header and return a Python dictionary of key:value pairs
    """

    header = {} # Initialise as a nil dictionary.

    for key, val in h5["data"].attrs.items():
        #if six.PY3:
        #    key = bytes(key, "ascii")
        if isinstance(val, bytes):
            val = val.decode("ascii")
        if key == "src_raj":
            header[key] = Angle(val, unit="hr")
        elif key == "src_dej":
            header[key] = Angle(val, unit="deg")
        else:
            header[key] = val

    return header


def examine(filename):
    r""" Diagnose the given HDF5 file"""
    h5 = h5py.File(filename, mode="r")
    examine_h5(h5)
    print("h5diag: data shape:", h5["data"].shape)


def cmd_tool(args=None):
    """ Command line tool h5diag """
    parser = ArgumentParser(description="Command line utility for diagnosing HDF5 files.")
    parser.add_argument("filename", type=str, help="Path of file to read")
    if args is None:
        args = sys.argv[1:]
    parse_args = parser.parse_args(args)

    if not os.path.isfile(parse_args.filename):
        oops("Not a file: {}".format(parse_args.filename))
    if not h5py.is_hdf5(parse_args.filename):
        oops("Not an HDF5 file: {}".format(parse_args.filename))

    print("h5diag: Begin")
    examine(parse_args.filename)
    print("h5diag: No errors detected")


if __name__ == "__main__":
    cmd_tool()



#!/usr/bin/env python
""" This model handles file types.
"""

import os
import h5py
import six

# This import relates to a failing project, so we will
# probably want to remove the import at some point in the future.
import blimpy.io.sigproc

# import pdb;# pdb.set_trace()

from blimpy.io.fil_reader import FilReader
from blimpy.io.hdf_reader import H5Reader

def open_file(filename, f_start=None, f_stop=None,t_start=None, t_stop=None,load_data=True,max_load=None):
    """Open a HDF5 or filterbank file

    Returns instance of a Reader to read data from file.

    ================== ==================================================
    Filename extension File type
    ================== ==================================================
    h5, hdf5           HDF5 format
    fil                fil format
    *other*            Will raise NotImplementedError
    ================== ==================================================

    """
    if not os.path.isfile(filename):
        type(filename)
        print(filename)
        raise IOError("No such file or directory: " + filename)

    filename = os.path.expandvars(os.path.expanduser(filename))
    # Get file extension to determine type
    ext = filename.split(".")[-1].strip().lower()

    if six.PY3:
        ext = bytes(ext, 'ascii')

    if h5py.is_hdf5(filename):
        # Open HDF5 file
        return H5Reader(filename, f_start=f_start, f_stop=f_stop, t_start=t_start, t_stop=t_stop,
                        load_data=load_data, max_load=max_load)
    if blimpy.io.sigproc.is_filterbank(filename):
        # Open FIL file
        return FilReader(filename, f_start=f_start, f_stop=f_stop, t_start=t_start, t_stop=t_stop, load_data=load_data, max_load=max_load)
    raise NotImplementedError('Cannot open this type of file with Waterfall: {}'.format(filename))



import pylab as plt
from .fluxcal import *
from .stokescal import *
from blimpy import Waterfall
import numpy as np

def get_diff(dio_cross,feedtype,**kwargs):
    """
    Returns ON-OFF for all Stokes parameters given a cross_pols noise diode measurement
    """
    #Get Stokes parameters, frequencies, and time sample length
    obs = Waterfall(dio_cross,max_load=150)
    freqs = obs.populate_freqs()
    tsamp = obs.header['tsamp']
    data = obs.data
    obs = None

    I,Q,U,V = get_stokes(data,feedtype)

    #Fold noise diode data
    I_OFF,I_ON = foldcal(I,tsamp,**kwargs)
    Q_OFF,Q_ON = foldcal(Q,tsamp,**kwargs)
    U_OFF,U_ON = foldcal(U,tsamp,**kwargs)
    V_OFF,V_ON = foldcal(V,tsamp,**kwargs)

    #Do ON-OFF subtraction
    Idiff = I_ON-I_OFF
    Qdiff = Q_ON-Q_OFF
    Udiff = U_ON-U_OFF
    Vdiff = V_ON-V_OFF

    return Idiff,Qdiff,Udiff,Vdiff,freqs

def plot_Stokes_diode(dio_cross,diff=True,feedtype='l',**kwargs):
    """
    Plots the uncalibrated full stokes spectrum of the noise diode.
    Use diff=False to plot both ON and OFF, or diff=True for ON-OFF
    """

    #If diff=True, get ON-OFF. If not get ON and OFF separately
    if diff==True:
        Idiff,Qdiff,Udiff,Vdiff,freqs = get_diff(dio_cross,feedtype,**kwargs)
    else:
        obs = Waterfall(dio_cross,max_load=150)
        freqs = obs.populate_freqs()
        tsamp = obs.header['tsamp']
        data = obs.data
        I,Q,U,V = get_stokes(data,feedtype)

        I_OFF,I_ON = foldcal(I,tsamp,**kwargs)
        Q_OFF,Q_ON = foldcal(Q,tsamp,**kwargs)
        U_OFF,U_ON = foldcal(U,tsamp,**kwargs)
        V_OFF,V_ON = foldcal(V,tsamp,**kwargs)

    #Plot spectra
    if diff==True:
        plt.plot(freqs,Idiff,'k-',label='I')
        plt.plot(freqs,Qdiff,'r-',label='Q')
        plt.plot(freqs,Udiff,'g-',label='U')
        plt.plot(freqs,Vdiff,'m-',label='V')
    else:
        plt.plot(freqs,I_ON,'k-',label='I ON')
        plt.plot(freqs,I_OFF,'k--',label='I OFF')
        plt.plot(freqs,Q_ON,'r-',label='Q ON')
        plt.plot(freqs,Q_OFF,'r--',label='Q OFF')
        plt.plot(freqs,U_ON,'g-',label='U ON')
        plt.plot(freqs,U_OFF,'g--',label='U OFF')
        plt.plot(freqs,V_ON,'m-',label='V ON')
        plt.plot(freqs,V_OFF,'m--',label='V OFF')

    plt.legend()
    plt.xlabel('Frequency (MHz)')
    plt.title('Uncalibrated Full Stokes Noise Diode Spectrum')
    plt.ylabel('Power (Counts)')


def plot_calibrated_diode(dio_cross,chan_per_coarse=8,feedtype='l',**kwargs):
    """
    Plots the corrected noise diode spectrum for a given noise diode measurement
    after application of the inverse Mueller matrix for the electronics chain.
    """
    #Get full stokes data for the ND observation
    obs = Waterfall(dio_cross,max_load=150)
    freqs = obs.populate_freqs()
    tsamp = obs.header['tsamp']
    data = obs.data
    obs = None
    I,Q,U,V = get_stokes(data,feedtype)
    data = None

    #Calculate Mueller Matrix variables for each coarse channel
    psis = phase_offsets(I,Q,U,V,tsamp,chan_per_coarse,feedtype,**kwargs)
    G = gain_offsets(I,Q,U,V,tsamp,chan_per_coarse,feedtype,**kwargs)

    #Apply the Mueller matrix to original noise diode data and refold
    I,Q,U,V = apply_Mueller(I,Q,U,V,G,psis,chan_per_coarse,feedtype)
    I_OFF,I_ON = foldcal(I,tsamp,**kwargs)
    Q_OFF,Q_ON = foldcal(Q,tsamp,**kwargs)
    U_OFF,U_ON = foldcal(U,tsamp,**kwargs)
    V_OFF,V_ON = foldcal(V,tsamp,**kwargs)

    #Delete data arrays for space
    I = None
    Q = None
    U = None
    V = None

    #Plot new ON-OFF spectra
    plt.plot(freqs,I_ON-I_OFF,'k-',label='I')
    plt.plot(freqs,Q_ON-Q_OFF,'r-',label='Q')
    plt.plot(freqs,U_ON-U_OFF,'g-',label='U')
    plt.plot(freqs,V_ON-V_OFF,'m-',label='V')

    plt.legend()
    plt.xlabel('Frequency (MHz)')
    plt.title('Calibrated Full Stokes Noise Diode Spectrum')
    plt.ylabel('Power (Counts)')


def plot_phase_offsets(dio_cross,chan_per_coarse=8,feedtype='l',ax1=None,ax2=None,legend=True,**kwargs):
    """
    Plots the calculated phase offsets of each coarse channel along with
    the UV (or QU) noise diode spectrum for comparison
    """
    #Get ON-OFF ND spectra
    Idiff,Qdiff,Udiff,Vdiff,freqs = get_diff(dio_cross,feedtype,**kwargs)
    obs = Waterfall(dio_cross,max_load=150)
    tsamp = obs.header['tsamp']
    data = obs.data
    obs = None
    I,Q,U,V = get_stokes(data,feedtype)

    #Get phase offsets and convert to degrees
    coarse_psis = phase_offsets(I,Q,U,V,tsamp,chan_per_coarse,feedtype,**kwargs)
    coarse_freqs = convert_to_coarse(freqs,chan_per_coarse)
    coarse_degs = np.degrees(coarse_psis)

    #Plot phase offsets
    if ax2==None:
        plt.subplot(211)
    else:
        axPsi = plt.axes(ax2)
        plt.setp(axPsi.get_xticklabels(),visible=False)
    plt.plot(coarse_freqs,coarse_degs,'ko',markersize=2,label='Coarse Channel $\psi$')
    plt.ylabel('Degrees')
    plt.grid(True)
    plt.title('Phase Offsets')
    if legend==True:
        plt.legend()

    #Plot U and V spectra
    if ax1==None:
        plt.subplot(212)
    else:
        axUV = plt.axes(ax1)
    plt.plot(freqs,Udiff,'g-',label='U')
    if feedtype=='l':
        plt.plot(freqs,Vdiff,'m-',label='V')
    if feedtype=='c':
        plt.plot(freqs,Qdiff,'r-',label='Q')
    plt.xlabel('Frequency (MHz)')
    plt.ylabel('Power (Counts)')
    plt.grid(True)
    if legend==True:
        plt.legend()

def plot_gain_offsets(dio_cross,dio_chan_per_coarse=8,feedtype='l',ax1=None,ax2=None,legend=True,**kwargs):
    """
    Plots the calculated gain offsets of each coarse channel along with
    the time averaged power spectra of the X and Y feeds
    """
    #Get ON-OFF ND spectra
    Idiff,Qdiff,Udiff,Vdiff,freqs = get_diff(dio_cross,feedtype,**kwargs)
    obs = Waterfall(dio_cross,max_load=150)
    tsamp = obs.header['tsamp']
    data = obs.data
    obs = None
    I,Q,U,V = get_stokes(data,feedtype)

    #Get phase offsets and convert to degrees
    coarse_G = gain_offsets(I,Q,U,V,tsamp,dio_chan_per_coarse,feedtype,**kwargs)
    coarse_freqs = convert_to_coarse(freqs,dio_chan_per_coarse)

    #Get X and Y spectra for the noise diode ON and OFF
    #If using circular feeds these correspond to LL and RR
    XX_OFF,XX_ON = foldcal(np.expand_dims(data[:,0,:],axis=1),tsamp,**kwargs)
    YY_OFF,YY_ON = foldcal(np.expand_dims(data[:,1,:],axis=1),tsamp,**kwargs)

    if ax1==None:
        plt.subplot(211)
    else:
        axG = plt.axes(ax1)
        plt.setp(axG.get_xticklabels(),visible=False)
    plt.plot(coarse_freqs,coarse_G,'ko',markersize=2)
    plt.ylabel(r'$\frac{\Delta G}{2}$',rotation=90)
    if feedtype=='l':
        plt.title('XY Gain Difference')
    if feedtype=='c':
        plt.title('LR Gain Difference')
    plt.grid(True)

    if ax2==None:
        plt.subplot(212)
    else:
        axXY = plt.axes(ax2,sharex=axG)
    if feedtype=='l':
        plt.plot(freqs,XX_OFF,'b-',label='XX')
        plt.plot(freqs,YY_OFF,'r-',label='YY')
    if feedtype=='c':
        plt.plot(freqs,XX_OFF,'b-',label='LL')
        plt.plot(freqs,YY_OFF,'r-',label='RR')
    plt.xlabel('Frequency (MHz)')
    plt.ylabel('Power (Counts)')
    if legend==True:
        plt.legend()

def plot_diode_fold(dio_cross,bothfeeds=True,feedtype='l',min_samp=-500,max_samp=7000,legend=True,**kwargs):
    """
    Plots the calculated average power and time sampling of ON (red) and
    OFF (blue) for a noise diode measurement over the observation time series
    """
    #Get full stokes data of ND measurement
    obs = Waterfall(dio_cross,max_load=150)
    tsamp = obs.header['tsamp']
    data = obs.data
    obs = None
    I,Q,U,V = get_stokes(data,feedtype)

    #Calculate time series, OFF and ON averages, and time samples for each
    tseriesI = np.squeeze(np.mean(I,axis=2))
    I_OFF,I_ON,OFFints,ONints = foldcal(I,tsamp,inds=True,**kwargs)
    if bothfeeds==True:
        if feedtype=='l':
            tseriesQ = np.squeeze(np.mean(Q,axis=2))
            tseriesX = (tseriesI+tseriesQ)/2
            tseriesY = (tseriesI-tseriesQ)/2
        if feedtype=='c':
            tseriesV = np.squeeze(np.mean(V,axis=2))
            tseriesR = (tseriesI+tseriesV)/2
            tseriesL = (tseriesI-tseriesV)/2

    stop = ONints[-1,1]

    #Plot time series and calculated average for ON and OFF
    if bothfeeds==False:
        plt.plot(tseriesI[0:stop],'k-',label='Total Power')
        for i in ONints:
            plt.plot(np.arange(i[0],i[1]),np.full((i[1]-i[0]),np.mean(I_ON)),'r-')
        for i in OFFints:
            plt.plot(np.arange(i[0],i[1]),np.full((i[1]-i[0]),np.mean(I_OFF)),'b-')
    else:
        if feedtype=='l':
            diff = np.mean(tseriesX)-np.mean(tseriesY)
            plt.plot(tseriesX[0:stop],'b-',label='XX')
            plt.plot(tseriesY[0:stop]+diff,'r-',label='YY (shifted)')
        if feedtype=='c':
            diff = np.mean(tseriesL)-np.mean(tseriesR)
            plt.plot(tseriesL[0:stop],'b-',label='LL')
            plt.plot(tseriesR[0:stop]+diff,'r-',label='RR (shifted)')

    #Calculate plotting limits
    if bothfeeds==False:
        lowlim = np.mean(I_OFF)-(np.mean(I_ON)-np.mean(I_OFF))/2
        hilim = np.mean(I_ON)+(np.mean(I_ON)-np.mean(I_OFF))/2
        plt.ylim((lowlim,hilim))

    plt.xlim((min_samp,max_samp))
    plt.xlabel('Time Sample Number')
    plt.ylabel('Power (Counts)')
    plt.title('Noise Diode Fold')
    if legend==True:
        plt.legend()


def plot_fullcalib(dio_cross,feedtype='l',**kwargs):
    """
    Generates and shows five plots: Uncalibrated diode, calibrated diode, fold information,
    phase offsets, and gain offsets for a noise diode measurement. Most useful diagnostic plot to
    make sure calibration proceeds correctly.
    """

    plt.figure("Multiple Calibration Plots", figsize=(12,9))
    left, width = 0.075,0.435
    bottom, height = 0.45,0.5
    width2 = 0.232
    bottom2, height2 = 0.115,0.0975

    rect_uncal = [left,bottom,width,height]
    rect_cal = [left+width+0.025,bottom,width,height]
    rect_fold = [left,bottom2,width2,0.22]
    rect_gain1 = [left+width2+0.1,bottom2,width2,height2]
    rect_phase1 = [left+width2*2+0.1*2,bottom2,width2,height2]
    rect_gain2 = [left+width2+0.1,bottom2+height2+0.025,width2,height2]
    rect_phase2 = [left+width2*2+0.1*2,bottom2+height2+0.025,width2,height2]

    #--------
    axFold = plt.axes(rect_fold)
    print('Plotting Diode Fold')
    plot_diode_fold(dio_cross,bothfeeds=False,feedtype=feedtype,min_samp=2000,max_samp=5500,legend=False,**kwargs)

    #--------
    print('Plotting Gain Offsets')
    plot_gain_offsets(dio_cross,feedtype=feedtype,ax1=rect_gain2,ax2=rect_gain1,legend=False,**kwargs)

    #--------
    print('Plotting Phase Offsets')
    plot_phase_offsets(dio_cross,feedtype=feedtype,ax1=rect_phase1,ax2=rect_phase2,legend=False,**kwargs)
    plt.ylabel('')

    #--------
    ax_uncal = plt.axes(rect_uncal)
    print('Plotting Uncalibrated Diode')
    plot_Stokes_diode(dio_cross,feedtype=feedtype,**kwargs)

    #--------
    ax_cal = plt.axes(rect_cal,sharey=ax_uncal)
    print('Plotting Calibrated Diode')
    plot_calibrated_diode(dio_cross,feedtype=feedtype,**kwargs)
    plt.ylabel('')
    plt.setp(ax_cal.get_yticklabels(),visible=False)

    plt.savefig(dio_cross[:-4]+'.stokescalib.png',dpi=2000)
    plt.show()

def plot_diodespec(ON_obs,OFF_obs,calflux,calfreq,spec_in,units='mJy',**kwargs):
    """
    Plots the full-band Stokes I spectrum of the noise diode (ON-OFF)
    """

    dspec = diode_spec(ON_obs,OFF_obs,calflux,calfreq,spec_in,**kwargs)
    obs = Waterfall(ON_obs,max_load=150)
    freqs = obs.populate_freqs()
    chan_per_coarse = obs.header['nchans']/obs.calc_n_coarse_chan()
    coarse_freqs = convert_to_coarse(freqs,chan_per_coarse)
    plt.ion()
    plt.figure()
    plt.plot(coarse_freqs,dspec)
    plt.xlabel('Frequency (MHz)')
    plt.ylabel('Flux Density ('+units+')')
    plt.title('Noise Diode Spectrum')



from blimpy import utils
import numpy as np
import pytest


def test_utils():
    assert utils.db(100) == 20.0
    assert utils.lin(20)  == 100.0
    assert utils.closest(np.array([0,1,2,3,4,5]), 2.2) == 2

def test_rebin():
    # 1D
    a = np.array([1, 1, 1, 1])
    aR = utils.rebin(a, 2)
    assert np.allclose(aR, np.array([1, 1]))

    # 2D
    b = np.array([[1,1,1,1], [2,2,2,2]])
    bR = utils.rebin(b, 1, 2)
    assert np.allclose(bR, [[1,1], [2,2]])
    bR = utils.rebin(b, None, 2)
    assert np.allclose(bR, [[1,1], [2,2]])
    bR = utils.rebin(b, 2, 1)
    assert np.allclose(bR, [1.5, 1.5, 1.5, 1.5])
    bR = utils.rebin(b, 2, None)
    assert np.allclose(bR, [1.5, 1.5, 1.5, 1.5])

    c = np.zeros([10, 10, 10])
    cR = utils.rebin(c, n_z=2)
    assert cR.shape == (10, 10, 5)
    cR = utils.rebin(c, n_y=2)
    assert cR.shape == (10, 5, 10)
    cR = utils.rebin(c, n_x=2)
    assert cR.shape == (5, 10, 10)

    c = np.zeros([10, 10, 10, 10])
    with pytest.raises(RuntimeError):
        utils.rebin(c, 2, 2)


if __name__ == "__main__":
    test_utils()
    test_rebin()


"""
# test_h52fil
"""

import pytest

import os
import blimpy as bl
from tests.data import voyager_h5


OUTDIR = os.path.dirname(voyager_h5) + "/"


def test_h52fil_conversion():
    """ Tests the conversion of fil files into h5 in both light and heavy modes.
    """

    # Creating test file.
    bl.h52fil.make_fil_file(voyager_h5, new_filename='test.fil')

    # Creating a "large" test file.
    bl.h52fil.make_fil_file(voyager_h5, new_filename='test_large.fil', max_load=0.001)

    # Testing filename
    bl.h52fil.make_fil_file(voyager_h5, new_filename='test')

    # Deleting test file
    os.remove('test.fil')
    os.remove('test_large.fil')

def test_help():
    """
    The user of these tests should verify that the help statement
    was printed as expected; this test merely verifies that the
    system exited.
    """
    with pytest.raises(SystemExit):
        bl.h52fil.cmd_tool(['-h'])

def test_cmd_tool():
    """
    This is the same Voyager test file, but now through the cmd tool.
    """
    bl.h52fil.cmd_tool([voyager_h5, '-n', OUTDIR + 'cmd.fil', '-l', '.001'])

def test_no_args():
    """
    The cmd tool needs to exit, mandating a file name.
    """
    with pytest.raises(SystemExit):
        bl.h52fil.cmd_tool(['-n', OUTDIR + 'cmd.fil', '-l', '.001'])

if __name__ == "__main__":
    test_h52fil_conversion()



r''' calcload.py - Calculate the Waterfall max_load value needed to load the data array for a given file.'''

import sys
from argparse import ArgumentParser
import numpy as np
import blimpy as bl


def calc_max_load(arg_path, verbose=False):
    r''' Calculate the max_load parameter value for a subsequent Waterfall instantiation.

    Algorithm:
        * A = minimum Waterfall object size.
        * B = data array size within one polarisation.
        * Return ceil(A + B in GB)
    '''
    wf = bl.Waterfall(arg_path, load_data=False)
    min_size = float(sys.getsizeof(wf.header)) + float(sys.getsizeof(wf))
    data_size = float(wf.header['nchans'] * wf.n_ints_in_file * wf.header['nbits']) / 8.0
    ngbytes = (min_size + data_size) / 1e9
    max_load = np.ceil(ngbytes)
    if verbose:
        print('calc_max_load: Waterfall object size excluding data = {}, data array size = {}, total GBs = {:.1f}'
              .format(min_size, data_size, ngbytes))
    return max_load


def cmd_tool(args=None):
    r'''Command line entrypoint for "calcload"'''
    p = ArgumentParser(description='Calculate the Waterfall max_load value needed to load the data array for a given file.')
    p.add_argument('filepath', type=str, help='Name of filepath to open (h5 or fil)')
    p.add_argument('-v', action='store_true', default=False, dest='verbose',
                       help='verbose output if True.')

    if args is None:
        args = p.parse_args()
    else:
        args = p.parse_args(args)

    gb = calc_max_load(args.filepath, args.verbose)
    if gb > 1.0:
        print('Use Waterfall instantiation with a max_load={}'.format(gb))
    else:
        print('Use Waterfall without a max_load= specification')


if __name__ == "__main__":
    cmd_tool()



try:
    from pyslalib import slalib as s
    HAS_SLALIB = True
except ImportError:
    HAS_SLALIB = False

import numpy as np
from astropy.coordinates import Angle

# Telescope coordinates (needed for LSR calc)
parkes_coords = (-32.998370, 148.263659,  324.00)
gbt_coords    = (38.4331294, 79.8398397, 824.36)



from blimpy import Waterfall
import numpy as np
from .fluxcal import foldcal

def get_stokes(cross_dat, feedtype='l'):
    '''Output stokes parameters (I,Q,U,V) for a rawspec
    cross polarization filterbank file'''

    #Compute Stokes Parameters
    if feedtype=='l':
        #I = XX+YY
        I = cross_dat[:,0,:]+cross_dat[:,1,:]
        #Q = XX-YY
        Q = cross_dat[:,0,:]-cross_dat[:,1,:]
        #U = 2*Re(XY)
        U = 2*cross_dat[:,2,:]
        #V = -2*Im(XY)
        V = -2*cross_dat[:,3,:]

    elif feedtype=='c':
        #I = LL+RR
        I = cross_dat[:,0,:]+cross_dat[:,1,:]
        #Q = 2*Re(RL)
        Q = 2*cross_dat[:,2,:]
        #U = 2*Im(RL)
        U = -2*cross_dat[:,3,:]
        #V = RR-LL
        V = cross_dat[:,1,:]-cross_dat[:,0,:]
    else:
        raise ValueError('feedtype must be \'l\' (linear) or \'c\' (circular)')

    #Add middle dimension to match Filterbank format
    I = np.expand_dims(I,axis=1)
    Q = np.expand_dims(Q,axis=1)
    U = np.expand_dims(U,axis=1)
    V = np.expand_dims(V,axis=1)

    #Compute linear polarization
    #L=np.sqrt(np.square(Q)+np.square(U))

    return I,Q,U,V

def convert_to_coarse(data,chan_per_coarse):
    '''
    Converts a data array with length n_chans to an array of length n_coarse_chans
    by averaging over the coarse channels
    '''
    #find number of coarse channels and reshape array
    num_coarse = data.size/chan_per_coarse
    data_shaped = np.array(np.reshape(data,(num_coarse,chan_per_coarse)))

    #Return the average over each coarse channel
    return np.mean(data_shaped[:,2:-1],axis=1)

def phase_offsets(Idat,Qdat,Udat,Vdat,tsamp,chan_per_coarse,feedtype='l',**kwargs):
    '''
    Calculates phase difference between X and Y feeds given U and V (U and Q for circular basis)
    data from a noise diode measurement on the target
    '''
    #Fold noise diode data and calculate ON OFF diferences for U and V
    if feedtype=='l':
        U_OFF,U_ON = foldcal(Udat,tsamp,**kwargs)
        V_OFF,V_ON = foldcal(Vdat,tsamp,**kwargs)
        Udiff = U_ON-U_OFF
        Vdiff = V_ON-V_OFF
        poffset = np.arctan2(-1*Vdiff,Udiff)

    if feedtype=='c':
        U_OFF,U_ON = foldcal(Udat,tsamp,**kwargs)
        Q_OFF,Q_ON = foldcal(Qdat,tsamp,**kwargs)
        Udiff = U_ON-U_OFF
        Qdiff = Q_ON-Q_OFF
        poffset = np.arctan2(Udiff,Qdiff)

    coarse_p =  convert_to_coarse(poffset,chan_per_coarse)

    #Correct for problems created by discontinuity in arctan
    #Find whether phase offsets have increasing or decreasing slope
    y = coarse_p[:6]
    x = np.arange(y.size)
    m = np.polyfit(x,y,1)[0]

    for i in range(coarse_p.size-3):
        if (m>0 and coarse_p[i+1]<coarse_p[i]) or (m<0 and coarse_p[i+1]>coarse_p[i]):
            coarse_p[i+1] = 2*coarse_p[i+2]-coarse_p[i+3]    #Move problem point near the next

    return coarse_p

def gain_offsets(Idat,Qdat,Udat,Vdat,tsamp,chan_per_coarse,feedtype='l',**kwargs):
    '''
    Determines relative gain error in the X and Y feeds for an
    observation given I and Q (I and V for circular basis) noise diode data.
    '''
    if feedtype=='l':
        #Fold noise diode data and calculate ON OFF differences for I and Q
        I_OFF,I_ON = foldcal(Idat,tsamp,**kwargs)
        Q_OFF,Q_ON = foldcal(Qdat,tsamp,**kwargs)

        #Calculate power in each feed for noise diode ON and OFF
        XX_OFF = (I_OFF+Q_OFF)/2
        YY_OFF = (I_OFF-Q_OFF)/2

        #Calculate gain offset (divided by 2) as defined in Heiles (2001)
        G = (XX_OFF-YY_OFF)/(XX_OFF+YY_OFF)

    if feedtype=='c':
        #Fold noise diode data and calculate ON OFF differences for I and Q
        I_OFF,I_ON = foldcal(Idat,tsamp,**kwargs)
        V_OFF,V_ON = foldcal(Vdat,tsamp,**kwargs)

        #Calculate power in each feed for noise diode ON and OFF
        RR_OFF = (I_OFF+V_OFF)/2
        LL_OFF = (I_OFF-V_OFF)/2

        #Calculate gain offset (divided by 2) as defined in Heiles (2001)
        G = (RR_OFF-LL_OFF)/(RR_OFF+LL_OFF)

    return convert_to_coarse(G,chan_per_coarse)

def apply_Mueller(I,Q,U,V, gain_offsets, phase_offsets, chan_per_coarse, feedtype='l'):
    '''
    Returns calibrated Stokes parameters for an observation given an array
    of differential gains and phase differences.
    '''

    #Find shape of data arrays and calculate number of coarse channels
    shape = I.shape
    ax0 = I.shape[0]
    ax1 = I.shape[1]
    nchans = I.shape[2]
    ncoarse = nchans/chan_per_coarse

    #Reshape data arrays to separate coarse channels
    I = np.reshape(I,(ax0,ax1,ncoarse,chan_per_coarse))
    Q = np.reshape(Q,(ax0,ax1,ncoarse,chan_per_coarse))
    U = np.reshape(U,(ax0,ax1,ncoarse,chan_per_coarse))
    V = np.reshape(V,(ax0,ax1,ncoarse,chan_per_coarse))

    #Swap axes 2 and 3 to in order for broadcasting to work correctly
    I = np.swapaxes(I,2,3)
    Q = np.swapaxes(Q,2,3)
    U = np.swapaxes(U,2,3)
    V = np.swapaxes(V,2,3)

    #Apply top left corner of electronics chain inverse Mueller matrix
    a = 1/(1-gain_offsets**2)
    if feedtype=='l':
        Icorr = a*(I-gain_offsets*Q)
        Qcorr = a*(-1*gain_offsets*I+Q)
        I = None
        Q = None
    if feedtype=='c':
        Icorr = a*(I-gain_offsets*V)
        Vcorr = a*(-1*gain_offsets*I+V)
        I = None
        V = None

    #Apply bottom right corner of electronics chain inverse Mueller matrix
    if feedtype=='l':
        Ucorr = U*np.cos(phase_offsets)-V*np.sin(phase_offsets)
        Vcorr = U*np.sin(phase_offsets)+V*np.cos(phase_offsets)
        U = None
        V = None
    if feedtype=='c':
        Qcorr = Q*np.cos(phase_offsets)+U*np.sin(phase_offsets)
        Ucorr = -1*Q*np.sin(phase_offsets)+U*np.cos(phase_offsets)
        Q = None
        U = None

    #Reshape arrays to original shape
    Icorr = np.reshape(np.swapaxes(Icorr,2,3),shape)
    Qcorr = np.reshape(np.swapaxes(Qcorr,2,3),shape)
    Ucorr = np.reshape(np.swapaxes(Ucorr,2,3),shape)
    Vcorr = np.reshape(np.swapaxes(Vcorr,2,3),shape)

    #Return corrected data arrays
    return Icorr,Qcorr,Ucorr,Vcorr

def calibrate_pols(cross_pols,diode_cross,obsI=None,onefile=True,feedtype='l',**kwargs):
    '''
    Write Stokes-calibrated filterbank file for a given observation
    with a calibrator noise diode measurement on the source

    Parameters
    ----------
    cross_pols : string
        Path to cross polarization filterbank file (rawspec output) for observation to be calibrated
    diode_cross : string
        Path to cross polarization filterbank file of noise diode measurement ON the target
    obsI : string
        Path to Stokes I filterbank file of main observation (only needed if onefile=False)
    onefile : boolean
        True writes all calibrated Stokes parameters to a single filterbank file,
        False writes four separate files
    feedtype : 'l' or 'c'
        Basis of antenna dipoles. 'c' for circular, 'l' for linear
    '''
    #Obtain time sample length, frequencies, and noise diode data
    obs = Waterfall(diode_cross,max_load=150)
    cross_dat = obs.data
    tsamp = obs.header['tsamp']

    #Calculate number of coarse channels in the noise diode measurement (usually 8)
    dio_ncoarse = obs.calc_n_coarse_chan()
    dio_nchans = obs.header['nchans']
    dio_chan_per_coarse = dio_nchans/dio_ncoarse
    obs = None
    Idat,Qdat,Udat,Vdat = get_stokes(cross_dat,feedtype)
    cross_dat = None
    #Calculate differential gain and phase from noise diode measurements
    print('Calculating Mueller Matrix variables')
    gams = gain_offsets(Idat,Qdat,Udat,Vdat,tsamp,dio_chan_per_coarse,feedtype,**kwargs)
    psis = phase_offsets(Idat,Qdat,Udat,Vdat,tsamp,dio_chan_per_coarse,feedtype,**kwargs)

    #Clear data arrays to save memory
    Idat = None
    Qdat = None
    Udat = None
    Vdat = None

    #Get corrected Stokes parameters
    print('Opening '+cross_pols)
    cross_obs = Waterfall(cross_pols,max_load=150)
    obs_ncoarse = cross_obs.calc_n_coarse_chan()
    obs_nchans = cross_obs.header['nchans']
    obs_chan_per_coarse = obs_nchans/obs_ncoarse

    print('Grabbing Stokes parameters')
    I,Q,U,V = get_stokes(cross_obs.data,feedtype)

    print('Applying Mueller Matrix')
    I,Q,U,V = apply_Mueller(I,Q,U,V,gams,psis,obs_chan_per_coarse,feedtype)

    #Use onefile (default) to produce one filterbank file containing all Stokes information
    if onefile:
        cross_obs.data[:,0,:] = np.squeeze(I)
        cross_obs.data[:,1,:] = np.squeeze(Q)
        cross_obs.data[:,2,:] = np.squeeze(U)
        cross_obs.data[:,3,:] = np.squeeze(V)
        cross_obs.write_to_fil(cross_pols[:-15]+'.SIQUV.polcal.fil')
        print('Calibrated Stokes parameters written to '+cross_pols[:-15]+'.SIQUV.polcal.fil')
        return

    # If onefile=False and obsI=None, abort
    assert obsI is not None

    #Write corrected Stokes parameters to four filterbank files if onefile==False
    obs = Waterfall(obsI,max_load=150)
    obs.data = I
    obs.write_to_fil(cross_pols[:-15]+'.SI.polcal.fil')   #assuming file is named *.cross_pols.fil
    print('Calibrated Stokes I written to '+cross_pols[:-15]+'.SI.polcal.fil')

    obs.data = Q
    obs.write_to_fil(cross_pols[:-15]+'.Q.polcal.fil')   #assuming file is named *.cross_pols.fil
    print('Calibrated Stokes Q written to '+cross_pols[:-15]+'.Q.polcal.fil')

    obs.data = U
    obs.write_to_fil(cross_pols[:-15]+'.U.polcal.fil')   #assuming file is named *.cross_pols.fil
    print('Calibrated Stokes U written to '+cross_pols[:-15]+'.U.polcal.fil')

    obs.data = V
    obs.write_to_fil(cross_pols[:-15]+'.V.polcal.fil')   #assuming file is named *.cross_pols.fil
    print('Calibrated Stokes V written to '+cross_pols[:-15]+'.V.polcal.fil')


def fracpols(cross_dat, **kwargs):
    '''Output fractional linear and circular polarizations for a
    rawspec cross polarization .fil file. NOT STANDARD USE'''

    I,Q,U,V,L=get_stokes(cross_dat, **kwargs)
    return L/I,V/I

def write_stokefils(cross_dat, str_I, Ifil=False, Qfil=False, Ufil=False, Vfil=False, Lfil=False, **kwargs):
    '''Writes up to 5 new filterbank files corresponding to each Stokes
    parameter (and total linear polarization L) for a given cross polarization .fil file'''

    I,Q,U,V,L=get_stokes(cross_dat, **kwargs)
    obs = Waterfall(str_I, max_load=150) #Load filterbank file to write stokes data to
    if Ifil:
        obs.data = I
        obs.write_to_fil(cross_dat[:-15]+'.I.fil')   #assuming file is named *.cross_pols.fil

    if Qfil:
        obs.data = Q
        obs.write_to_fil(cross_dat[:-15]+'.Q.fil')   #assuming file is named *.cross_pols.fil

    if Ufil:
        obs.data = U
        obs.write_to_fil(cross_dat[:-15]+'.U.fil')   #assuming file is named *.cross_pols.fil

    if Vfil:
        obs.data = V
        obs.write_to_fil(cross_dat[:-15]+'.V.fil')   #assuming file is named *.cross_pols.fil

    if Lfil:
        obs.data = L
        obs.write_to_fil(cross_dat[:-15]+'.L.fil')   #assuming file is named *.cross_pols.fil


def write_polfils(cross_dat, str_I, **kwargs):
    '''Writes two new filterbank files containing fractional linear and
    circular polarization data'''

    lin,circ=fracpols(cross_dat, **kwargs)
    obs = Waterfall(str_I, max_load=150)

    obs.data = lin
    obs.write_to_fil(cross_dat[:-15]+'.linpol.fil')   #assuming file is named *.cross_pols.fil

    obs.data = circ
    obs.write_to_fil(cross_dat[:-15]+'.circpol.fil')   #assuming file is named *.cross_pols.fil



import time
import h5py
import hdf5plugin
from blimpy import utils


def write_to_hdf5(wf, filename_out, f_scrunch=None, *args, **kwargs):
    """ Write data to HDF5 file.
        It check the file size then decides how to write the file.

    Args:
        filename_out (str): Name of output file
        f_scrunch (int or None): Average (scrunch) N channels together
    """

    #For timing how long it takes to write a file.
    t0 = time.time()

    #Update header
    wf._update_header()

    if wf.container.isheavy():
        __write_to_hdf5_heavy(wf, filename_out, f_scrunch=f_scrunch, *args, **kwargs)
    else:
        __write_to_hdf5_light(wf, filename_out, f_scrunch=f_scrunch, *args, **kwargs)

    t1 = time.time()
    wf.logger.info('Conversion time: %2.2fsec' % (t1- t0))


def __write_to_hdf5_heavy(wf, filename_out, f_scrunch=None, *args, **kwargs):
    """ Write data to HDF5 file.

    Args:
        filename_out (str): Name of output file
        f_scrunch (int or None): Average (scrunch) N channels together
    """

    block_size = 0

    # Note that a chunk is not a blob!!
    # chunk_dim = wf._get_chunk_dimensions() <-- seems intended for raw to fil
    # And, chunk dimensions should not exceed the Waterfall selection shape dimensions.
    chunk_list = list(wf._get_chunk_dimensions())
    for ix in range(0, len(chunk_list)):
        if chunk_list[ix] > wf.selection_shape[ix]:
            chunk_list[ix] = wf.selection_shape[ix]
    chunk_dim = tuple(chunk_list)
    blob_dim  = wf._get_blob_dimensions(chunk_dim)
    n_blobs   = wf.container.calc_n_blobs(blob_dim)

    with h5py.File(filename_out, 'w') as h5:

        h5.attrs['CLASS'] = 'FILTERBANK'
        h5.attrs['VERSION'] = '1.0'

        bs_compression = hdf5plugin.Bitshuffle(nelems=0, lz4=True)['compression']
        bs_compression_opts = hdf5plugin.Bitshuffle(nelems=0, lz4=True)['compression_opts']

        dout_shape     = list(wf.selection_shape)    # Make sure not a tuple
        dout_chunk_dim = list(chunk_dim)

        if f_scrunch is not None:
            dout_shape[-1] //= f_scrunch
            dout_chunk_dim[-1] //= f_scrunch
            wf.header['foff'] *= f_scrunch

        dset = h5.create_dataset('data',
                                 shape=tuple(dout_shape),
                                 chunks=tuple(dout_chunk_dim),
                                 compression=bs_compression,
                                 compression_opts=bs_compression_opts,
                                 dtype=wf.data.dtype)

        dset_mask = h5.create_dataset('mask',
                                      shape=tuple(dout_shape),
                                      chunks=tuple(dout_chunk_dim),
                                      compression=bs_compression,
                                      compression_opts=bs_compression_opts,
                                      dtype='uint8')

        dset.dims[2].label = b"frequency"
        dset.dims[1].label = b"feed_id"
        dset.dims[0].label = b"time"

        dset_mask.dims[2].label = b"frequency"
        dset_mask.dims[1].label = b"feed_id"
        dset_mask.dims[0].label = b"time"

        # Copy over header information as attributes
        for key, value in wf.header.items():
            dset.attrs[key] = value

        if blob_dim[wf.freq_axis] < wf.selection_shape[wf.freq_axis]:

            wf.logger.info('Using %i n_blobs to write the data.'% n_blobs)
            for ii in range(0, n_blobs):
                wf.logger.info('Reading %i of %i' % (ii + 1, n_blobs))

                bob = wf.container.read_blob(blob_dim, n_blob=ii)

                #-----
                #Using channels instead of frequency.
                c_start = wf.container.chan_start_idx + ii * blob_dim[wf.freq_axis]
                t_start = wf.container.t_start + (c_start / wf.selection_shape[wf.freq_axis]) * blob_dim[wf.time_axis]
                t_stop = t_start + blob_dim[wf.time_axis]

                # Reverse array if frequency axis is flipped
#                     if self.header['foff'] < 0:
#                         c_stop = self.selection_shape[self.freq_axis] - (c_start)%self.selection_shape[self.freq_axis]
#                         c_start = c_stop - blob_dim[self.freq_axis]
#                     else:
                c_start = (c_start) % wf.selection_shape[wf.freq_axis]
                c_stop = c_start + blob_dim[wf.freq_axis]
                #-----

                if f_scrunch is not None:
                    c_start //= f_scrunch
                    c_stop  //= f_scrunch
                    bob = utils.rebin(bob, n_z=f_scrunch)

                wf.logger.debug(t_start,t_stop,c_start,c_stop)
                dset[t_start:t_stop,0,c_start:c_stop] = bob[:]

        else:

            wf.logger.info('Using %i n_blobs to write the data.'% n_blobs)
            for ii in range(0, n_blobs):
                wf.logger.info('Reading %i of %i' % (ii + 1, n_blobs))

                bob = wf.container.read_blob(blob_dim, n_blob=ii)
                t_start = wf.container.t_start + ii * blob_dim[wf.time_axis]

                #This prevents issues when the last blob is smaller than the others in time
                if (ii+1)*blob_dim[wf.time_axis] > wf.n_ints_in_file:
                    t_stop = wf.n_ints_in_file
                else:
                    t_stop = (ii+1)*blob_dim[wf.time_axis]

                if f_scrunch is not None:
                    bob = utils.rebin(bob, n_z=f_scrunch)

                dset[t_start:t_stop] = bob[:]


def __write_to_hdf5_light(wf, filename_out, f_scrunch=None, *args, **kwargs):
    """ Write data to HDF5 file in one go.

    Args:
        filename_out (str): Name of output file
        f_scrunch (int or None): Average (scrunch) N channels together
    """

    block_size = 0

    with h5py.File(filename_out, 'w') as h5:

        h5.attrs['CLASS']   = 'FILTERBANK'
        h5.attrs['VERSION'] = '1.0'

        bs_compression = hdf5plugin.Bitshuffle(nelems=0, lz4=True)['compression']
        bs_compression_opts = hdf5plugin.Bitshuffle(nelems=0, lz4=True)['compression_opts']

        if f_scrunch is None:
            data_out = wf.data
        else:
            wf.logger.info('Frequency scrunching by %i' % f_scrunch)
            data_out = utils.rebin(wf.data, n_z=f_scrunch)
            wf.header['foff'] *= f_scrunch

        dset = h5.create_dataset('data',
                                 data=data_out,
                                 compression=bs_compression,
                                 compression_opts=bs_compression_opts)

        dset_mask = h5.create_dataset('mask',
                                      shape=data_out.shape,
                                      compression=bs_compression,
                                      compression_opts=bs_compression_opts,
                                      dtype='uint8')

        dset.dims[2].label = b"frequency"
        dset.dims[1].label = b"feed_id"
        dset.dims[0].label = b"time"

        dset_mask.dims[2].label = b"frequency"
        dset_mask.dims[1].label = b"feed_id"
        dset_mask.dims[0].label = b"time"

        # Copy over header information as attributes
        for key, value in wf.header.items():
            dset.attrs[key] = value




from .config import *

def compute_lst(wf):
    """ Compute LST for observation

    Computes local sidereal time (LST) for the observation, using SLALIB.

    Args:
        wf (bl.Waterfall): blimpy Waterfall object.
    """
    if wf.header['telescope_id'] == 6:
        wf.coords = gbt_coords
    elif wf.header['telescope_id'] == 4:
        wf.coords = parkes_coords
    else:
        raise RuntimeError("Currently only Parkes and GBT supported")
    if HAS_SLALIB:
        # dut1 = (0.2 /3600.0) * np.pi/12.0
        dut1 = 0.0
        mjd = wf.header['tstart']
        tellong = np.deg2rad(wf.coords[1])
        last = s.sla_gmst(mjd) - tellong + s.sla_eqeqx(mjd) + dut1
        # lmst = s.sla_gmst(mjd) - tellong
        if last < 0.0: last = last + 2.0 * np.pi
        return last
    else:
        raise RuntimeError("This method requires pySLALIB")



r'''Reader for Filterbank files (.fil)'''

import os

import numpy as np

from blimpy.io import sigproc
from blimpy.io.base_reader import Reader, logger, GIGA


class FilReader(Reader):
    """ This class handles .fil files.
    """

    def __init__(self, filename,f_start=None, f_stop=None,t_start=None, t_stop=None, load_data=True, max_load=None):
        """ Constructor.

        Args:
            filename (str): filename of blimpy file.
            f_start (float): start frequency, in MHz
            f_stop (float): stop frequency, in MHz
            t_start (int): start time bin
            t_stop (int): stop time bin
            max_load (float): memory limit in gigabytes
        """
        super(FilReader, self).__init__()

        self.header_keywords_types = sigproc.header_keyword_types

        if filename and os.path.isfile(filename):
            self.filename = filename
            self.load_data = load_data
            self.header = self.read_header()
            self.file_size_bytes = os.path.getsize(self.filename)
            self.idx_data = sigproc.len_header(self.filename)
            self.n_channels_in_file  = self.header['nchans']
            self.n_beams_in_file = self.header['nifs'] #Placeholder for future development.
            self.n_pols_in_file = 1 #Placeholder for future development.
            self._n_bytes = int(self.header['nbits'] / 8)  #number of bytes per digit.
            self._d_type = self._setup_dtype()
            self._setup_n_ints_in_file()
            self.file_shape = (self.n_ints_in_file,self.n_beams_in_file,self.n_channels_in_file)

            if self.header['foff'] < 0:
                self.f_end  = self.header['fch1']
                self.f_begin  = self.f_end + self.n_channels_in_file*self.header['foff']
            else:
                self.f_begin  = self.header['fch1']
                self.f_end  = self.f_begin + self.n_channels_in_file*self.header['foff']

            self.t_begin = 0
            self.t_end = self.n_ints_in_file

            #Taking care all the frequencies are assigned correctly.
            self._setup_selection_range(f_start=f_start, f_stop=f_stop, t_start=t_start, t_stop=t_stop, init=True)
            #Convert input frequencies into what their corresponding channel number would be.
            self._setup_chans()
            #Update frequencies ranges from channel number.
            self._setup_freqs()

            self.freq_axis = 2
            self.time_axis = 0
            self.beam_axis = 1  # Place holder

#EE ie.
#           spec = np.squeeze(fil_file.data)
            # set start of data, at real length of header  (future development.)
#            self.datastart=self.hdrraw.find('HEADER_END')+len('HEADER_END')+self.startsample*self.channels

            #Applying data size limit to load.
            if max_load is not None and max_load > 0:
                self.max_data_array_size = max_load * GIGA

            if self.file_size_bytes > self.max_data_array_size:
                self.large_file = True
            else:
                self.large_file = False

            if self.load_data:
                if self.large_file:
                    if self.f_start or self.f_stop or self.t_start or self.t_stop:
                        if self.isheavy():
                            self.warn_memory("Selection", self._calc_selection_size())
                            self._init_empty_selection()
                        else:
                            self.read_data()
                    else:
                        self.warn_memory("File", self.file_size_bytes)
                        self._init_empty_selection()
                else:
                    self.read_data()
            else:
                logger.debug("Skipping loading data ...")
                self._init_empty_selection()
        else:
            raise IOError("Need a file to open, please give me one!")

    def _setup_n_ints_in_file(self):
        """ Calculate the number of integrations in the file. """
        self.n_ints_in_file = sigproc.calc_n_ints_in_file(self.filename)


    def read_header(self, return_idxs=False):
        """ Read blimpy header and return a Python dictionary of key:value pairs

        Args:
            filename (str): name of file to open

        Optional args:
            return_idxs (bool): Default False. If true, returns the file offset indexes
                                for values

        Returns:
            Python dict of key:value pairs, OR returns file offset indexes for values.

        """
        self.header = sigproc.read_header(self.filename, return_idxs=return_idxs)
        return self.header

    def read_data(self, f_start=None, f_stop=None,t_start=None, t_stop=None):
        """ Read data.
        """

        self._setup_selection_range(f_start=f_start, f_stop=f_stop, t_start=t_start, t_stop=t_stop)

        #check if selection is small enough.
        if self.isheavy():
            self.warn_memory("Selection", self._calc_selection_size())
            self.data = np.array([0], dtype=self._d_type)
            return None

        #Convert input frequencies into what their corresponding channel number would be.
        self._setup_chans()
        #Update frequencies ranges from channel number.
        self._setup_freqs()

        n_chans = self.header['nchans']
        n_chans_selected = self.selection_shape[self.freq_axis]
        n_ifs   = self.header['nifs']

        # Load binary data
        f = open(self.filename, 'rb')
        f.seek(int(self.idx_data))

        # now check to see how many integrations requested
        n_ints = self.t_stop - self.t_start

        # Seek to first integration
        f.seek(int(self.t_start * self._n_bytes  * n_ifs * n_chans), 1)

        #Loading  data
        self.data = np.zeros((n_ints, n_ifs, n_chans_selected), dtype=self._d_type)

        for ii in range(n_ints):
            for jj in range(n_ifs):
                f.seek(int(self._n_bytes  * self.chan_start_idx), 1) # 1 = from current location
                dd = np.fromfile(f, count=n_chans_selected, dtype=self._d_type)

                # Reverse array if frequency axis is flipped
#                     if self.header['foff'] < 0:
#                         dd = dd[::-1]

                self.data[ii, jj] = dd

                f.seek(int(self._n_bytes  * (n_chans - self.chan_stop_idx)), 1)  # Seek to start of next block

        # Give the FD back to the O/S.
        f.close()

    def _find_blob_start(self):
        """Find first blob from selection.
        """

        # Convert input frequencies into what their corresponding channel number would be.
        self._setup_chans()

        # Check which is the blob time offset
        blob_time_start = self.t_start

        # Check which is the blob frequency offset (in channels)
        blob_freq_start = self.chan_start_idx

        blob_start = blob_time_start * self.n_channels_in_file + blob_freq_start

        return blob_start

    def read_blob(self,blob_dim,n_blob=0):
        """Read blob from a selection.
        """

        n_blobs = self.calc_n_blobs(blob_dim)
        if n_blob > n_blobs or n_blob < 0:
            raise ValueError('Please provide correct n_blob value. Given %i, but max values is %i'%(n_blob,n_blobs))

        # This prevents issues when the last blob is smaller than the others in time.
        if blob_dim[self.time_axis]*(n_blob+1) > self.selection_shape[self.time_axis]:
            updated_blob_dim = (int(self.selection_shape[self.time_axis] - blob_dim[self.time_axis]*n_blob), 1, int(blob_dim[self.freq_axis]))
        else:
            updated_blob_dim = [int(i) for i in blob_dim]

        blob_start = self._find_blob_start()
        blob = np.zeros(updated_blob_dim, dtype=self._d_type)

        # EE: For now; also assuming one polarization and one beam.

        # Assuming the blob will loop over the whole frequency range.
        if self.f_start == self.f_begin and self.f_stop == self.f_end:

            blob_flat_size = np.prod(blob_dim)
            updated_blob_flat_size = np.prod(updated_blob_dim)

            # Load binary data
            with open(self.filename, 'rb') as f:
                f.seek(int(self.idx_data + self._n_bytes  * (blob_start + n_blob*blob_flat_size)))
                dd = np.fromfile(f, count=updated_blob_flat_size, dtype=self._d_type)

            if dd.shape[0] == updated_blob_flat_size:
                blob = dd.reshape(updated_blob_dim)
            else:
                logger.info('DD shape != blob shape.')
                blob = dd.reshape((int(dd.shape[0]/blob_dim[self.freq_axis]),blob_dim[self.beam_axis],blob_dim[self.freq_axis]))
        else:

            for blobt in range(updated_blob_dim[self.time_axis]):

                #Load binary data
                with open(self.filename, 'rb') as f:
                    f.seek(int(self.idx_data + self._n_bytes * (blob_start + n_blob*blob_dim[self.time_axis]*self.n_channels_in_file + blobt*self.n_channels_in_file)))
                    dd = np.fromfile(f, count=blob_dim[self.freq_axis], dtype=self._d_type)

                blob[blobt] = dd

#         if self.header['foff'] < 0:
#             blob = blob[:,:,::-1]

        return blob

    def read_all(self,reverse=True):
        """ read all the data.
            If reverse=True the x axis is flipped.
        """
        raise NotImplementedError('To be implemented')

        # go to start of the data
        self.filfile.seek(int(self.datastart))
        # read data into 2-D numpy array
#        data=np.fromfile(self.filfile,dtype=self.dtype).reshape(self.channels,self.blocksize,order='F')
        data=np.fromfile(self.filfile,dtype=self.dtype).reshape(self.blocksize, self.channels)
        if reverse:
            data = data[:,::-1]
        return data

    def read_row(self,rownumber,reverse=True):
        """ Read a block of data. The number of samples per row is set in self.channels
            If reverse=True the x axis is flipped.
        """
        raise NotImplementedError('To be implemented')

        # go to start of the row
        self.filfile.seek(int(self.datastart+self.channels*rownumber*(int(self.nbits/8))))
        # read data into 2-D numpy array
        data=np.fromfile(self.filfile,count=self.channels,dtype=self.dtype).reshape(1, self.channels)
        if reverse:
            data = data[:,::-1]
        return data

    def read_rows(self,rownumber,n_rows,reverse=True):
        """ Read a block of data. The number of samples per row is set in self.channels
            If reverse=True the x axis is flipped.
        """
        raise NotImplementedError('To be implemented')

        # go to start of the row
        self.filfile.seek(int(self.datastart+self.channels*rownumber*(int(self.nbits/8))))
        # read data into 2-D numpy array
        data=np.fromfile(self.filfile,count=self.channels*n_rows,dtype=self.dtype).reshape(n_rows, self.channels)
        if reverse:
            data = data[:,::-1]
        return data



import os
import numpy as np

# Check if $DISPLAY is set (for handling plotting on remote machines with no X-forwarding)
import matplotlib

if 'DISPLAY' in os.environ.keys():
    import pylab as plt
else:
    matplotlib.use('Agg')
    import pylab as plt

from matplotlib.ticker import NullFormatter

plt.rcParams['axes.formatter.useoffset'] = False


MAX_PLT_POINTS      = 65536                  # Max number of points in matplotlib plot
MAX_IMSHOW_POINTS   = (8192, 4096)           # Max number of points in imshow plot


from .config import *
from ..utils import rebin, db
from .plot_utils import calc_extent


def plot_waterfall(wf, f_start=None, f_stop=None, if_id=0, logged=True, cb=True, MJD_time=False, **kwargs):
    """ Plot waterfall of data

    Args:
        f_start (float): start frequency, in MHz
        f_stop (float): stop frequency, in MHz
        logged (bool): Plot in linear (False) or dB units (True),
        cb (bool): for plotting the colorbar
        kwargs: keyword args to be passed to matplotlib imshow()
    """

    plot_f, plot_data = wf.grab_data(f_start, f_stop, if_id)
    
    # imshow does not support int8, so convert to floating point
    plot_data = plot_data.astype('float32')
    
    # Using accending frequency for all plots.
    if wf.header['foff'] < 0:
        plot_data = plot_data[..., ::-1]  # Reverse data
        plot_f = plot_f[::-1]

    if logged:
        plot_data = db(plot_data)

    # Make sure waterfall plot is under 4k*4k
    dec_fac_x, dec_fac_y = 1, 1
    if plot_data.shape[0] > MAX_IMSHOW_POINTS[0]:
        dec_fac_x = int(plot_data.shape[0] / MAX_IMSHOW_POINTS[0])

    if plot_data.shape[1] > MAX_IMSHOW_POINTS[1]:
        dec_fac_y = int(plot_data.shape[1] / MAX_IMSHOW_POINTS[1])

    plot_data = rebin(plot_data, dec_fac_x, dec_fac_y)

    try:
        plt.title(wf.header['source_name'])
    except KeyError:
        plt.title(wf.filename)

    extent = calc_extent(wf, plot_f=plot_f, plot_t=wf.timestamps, MJD_time=MJD_time)

    plt.imshow(plot_data,
               aspect='auto',
               origin='lower',
               rasterized=True,
               interpolation='nearest',
               extent=extent,
               cmap='viridis',
               **kwargs
               )
    if cb:
        plt.colorbar()
    plt.xlabel("Frequency [MHz]")
    if MJD_time:
        plt.ylabel("Time [MJD]")
    else:
        plt.ylabel("Time [s]")



import blimpy as bl
from tests.data import voyager_fil

""" This test is broken for some reason.
It works on my machine but fails the Docker Travis CI,
so I will keep working on it.

def test_size():
    bl.match_fils.find_header_size(voyager_fil)
"""

def test_reset_works():
    """
    The test is trivial, but coverage must be
    increased at all costs.
    """
    a, b = bl.match_fils.reset_outs()
    assert a is None
    assert b is None

def test_batcher():
    """
    This test may be obsolete once/if test_cmd
        can be patched up.
    """
    bl.match_fils.make_batch_script()



import blimpy as bl
import numpy as np
from pprint import pprint
import pylab as plt

from tests.data import voyager_fil, voyager_h5
from blimpy.plotting import plot_waterfall, plot_spectrum, plot_spectrum_min_max, \
    plot_kurtosis, plot_time_series, plot_all


def test_plot_waterfall():
    """ Load Voyager dataset and test plotting """

    a = bl.Waterfall(voyager_h5)

    plt.figure("TEST PLOTTING", figsize=(10, 8))
    plt.subplot(3, 2, 1)
    plot_waterfall(a)

    plt.subplot(3, 2, 2)
    plot_spectrum(a)

    plt.subplot(3, 2, 3)
    plot_spectrum_min_max(a)

    plt.subplot(3, 2, 4)
    plot_kurtosis(a)

    plt.subplot(3, 2, 5)
    plot_time_series(a)

    plt.tight_layout()
    plt.savefig("test_plotting.png")

    plt.figure("TEST PLOT_ALL", figsize=(10, 8))
    plot_all(a)
    plt.savefig("test_plotting_plot_all.png")


def test_plot_waterfall_classmethod():
    """ Load Voyager dataset and test plotting """

    a = bl.Waterfall(voyager_h5)

    plt.figure("TEST PLOTTING CLASS", figsize=(10, 8))
    plt.subplot(3, 2, 1)
    a.plot_waterfall()

    plt.subplot(3, 2, 2)
    a.plot_spectrum()

    plt.subplot(3, 2, 3)
    a.plot_spectrum_min_max()

    plt.subplot(3, 2, 4)
    a.plot_kurtosis()

    plt.subplot(3, 2, 5)
    a.plot_time_series()
    plt.tight_layout()

    plt.savefig("test_plotting_classmethod.png")

    plt.figure("TEST PLOT_ALL CLASS", figsize=(10, 8))
    a.plot_all()
    plt.savefig("test_plotting_plot_all_classmethod.png")


if __name__ == "__main__":
    test_plot_waterfall()
    test_plot_waterfall_classmethod()


from blimpy.utils import unpack_1to8, unpack_2to8, unpack_4to8, unpack
import numpy as np
import pytest

def test_1to8():
    a = np.array([0b01010101, 0b10101010], dtype=np.uint8)
    b = np.array([0,1,0,1,0,1,0,1,1,0,1,0,1,0,1,0])
    c = unpack_1to8(a)
    assert np.allclose(b, c)
    print(b)
    print(c)

def test_2to8():
    # Create an array that should come out as [0, 1, 2, 3, 3, 2, 1, 0]
    # In binary, this is [00, 01, 10, 11, 11, 10, 01, 00]
    # Convert to 8-bit, this is [0b00011011, 0b11100100]

    a = np.array([0b00011011, 0b11100100], dtype=np.uint8)
    b = np.array([40, 12, -12, -40, -40, -12, 12, 40], dtype=np.int8)

    c = unpack_2to8(a)

    assert np.allclose(b, c)

def test_4to8():
    # Create an array that should come out as [0, 1, 2, 3, 3, 2, 1, 0]
    # In binary, this is [00, 01, 10, 11, 11, 10, 01, 00]
    # Convert to 8-bit, this is [0b00011011, 0b11100100]

    # Test 4-bit unpack
    a = np.array([0b00000001, 0b00100011], dtype=np.uint8)
    b = np.array([0, 1, 2, 3], dtype=np.uint8)
    c = unpack_4to8(a)
    assert np.allclose(b, c)

def test_unpack():

    # Test 2-bit unpack
    a = np.array([0b00011011, 0b11100100], dtype=np.uint8)
    b = np.array([40, 12, -12, -40, -40, -12, 12, 40], dtype=np.int8)
    c = unpack(a, 2)
    assert np.allclose(b, c)

    # Catch exceptions
    with pytest.raises(ValueError):
        unpack(a, 16)  # nbit <= 8 is reqd
    with pytest.raises(ValueError):
        unpack(a, 3)  # nbit must divide 8 (1,2,4 or 8)
    z = np.array([1,2,3], dtype='float32')
    with pytest.raises(TypeError):
        unpack(z, 2) # input data must be 8-bit

    # Test 4-bit unpack
    a = np.array([0b00000001, 0b00100011], dtype=np.uint8)
    b = np.array([0, 1, 2, 3], dtype=np.uint8)
    c = unpack(a, 4)
    print(b)
    print(c)
    assert np.allclose(b, c)

    # Test 1-bit unpack
    a = np.array([0b01010101, 0b10101010], dtype=np.uint8)
    b = np.array([0,1,0,1,0,1,0,1,1,0,1,0,1,0,1,0])
    c = unpack(a, 1)
    print(b)
    print(c)
    assert np.allclose(b, c)

    # Test 8-bit!
    c = unpack(a, 8)
    assert np.allclose(a, c)

if __name__ == "__main__":
    test_1to8()
    test_2to8()
    test_4to8()
    test_unpack()


#!/usr/bin/env python
"""
    Simple script for quickly making a .fil file from a .h5.

    ..author: Emilio Enriquez (jeenriquez@gmail.com)

    July 28th 2017
"""

try:
    from .waterfall import Waterfall
except:
    from waterfall import Waterfall

from optparse import OptionParser
import sys
import os

#------
# Logging set up
import logging
logger = logging.getLogger(__name__)

level_log = logging.INFO

if level_log == logging.INFO:
    stream = sys.stdout
    format = '%(name)-15s %(levelname)-8s %(message)s'
else:
    stream =  sys.stderr
    format = '%%(relativeCreated)5d (name)-15s %(levelname)-8s %(message)s'

logging.basicConfig(format=format,stream=stream,level = level_log)
#------


def make_fil_file(filename,out_dir='./', new_filename=None, max_load = None):
    """ Converts file to Sigproc filterbank (.fil) format.  Default saves output in current dir.
    """

    fil_file = Waterfall(filename, max_load = max_load)
    if not new_filename:
        new_filename = out_dir+filename.replace('.h5','.fil').split('/')[-1]

    if '.fil' not in new_filename:
        new_filename = new_filename+'.fil'

    fil_file.write_to_fil(new_filename)

def cmd_tool(flags=None):
    """  Command line utility for converting HDF5 (.h5) to Sigproc filterbank (.fil) format

    Usage:
        h52fil <FULL_PATH_TO_FIL_FILE> [options]

    Options:
        -h, --help            show this help message and exit
        -o OUT_DIR, --out_dir=OUT_DIR
                              Location for output files. Default: local dir.
        -n NEW_FILENAME, --new_filename=NEW_FILENAME
                              New filename. Default: replaces extension to .fil
        -d, --delete_input    This option deletes the input file after conversion.
        -l MAX_LOAD           Maximum data limit to load. Default:1GB
    """

    p = OptionParser()
    p.set_usage('Command line utility for converting HDF5 (.h5) to Sigproc filterbank (.fil) format \n >>h52fil <FULL_PATH_TO_FIL_FILE> [options]')
    p.add_option('-o', '--out_dir', dest='out_dir', type='str', default='./',
                 help='Location for output files. Default: local dir. ')
    p.add_option('-n', '--new_filename', dest='new_filename', type='str', default='',
                 help='New filename. Default: replaces extension to .fil')
    p.add_option('-d', '--delete_input', dest='delete_input', action='store_true', default=False,
                 help='This option deletes the input file after conversion.')
    p.add_option('-l', action='store', default=None, dest='max_load', type=float,
                 help='Maximum data limit to load. Default:1GB')
    if flags is None:
        opts, args = p.parse_args(sys.argv[1:])
    else:
        opts, args = p.parse_args(flags)

    if len(args) != 1:
        logger.info('Please specify a file name \nExiting.')
        sys.exit()
    else:
        filename = args[0]

    make_fil_file(filename, out_dir = opts.out_dir, new_filename=opts.new_filename, max_load = opts.max_load)

    if opts.delete_input:
        logger.info("'Deleting input file: %s"%(filename))
        os.remove(filename)

if __name__ == "__main__":

    cmd_tool()



r"""
test_observatory.py
"""

from blimpy.ephemeris import Observatory


def error_msg(arg_string):
    r""" Just making clearer error messages """
    return "test_observatory.py: " + arg_string


def test_observatory_construction():
    r""" Constructor test """
    print("\n===== Begin test_observatory_construction")
    obs = Observatory()
    assert obs.get_telescope_name() == "Fake", error_msg("Wrong name for the fake observatory")
    obs = Observatory(telescope_id=4)
    assert obs.get_telescope_name() == "PARKES", error_msg("Wrong name for the Parkes observatory")
    assert obs.get_telescope_name_short() == "PK", \
        error_msg("Wrong short name for the Parkes observatory")
    obs = Observatory(telescope_name="GBT")
    assert obs.get_telescope_id() == 6, error_msg("Wrong Sigproc ID for the GBT observatory")
    print("===== End test_observatory_construction")


def test_observatory_values():
    r""" Observatory values test along with beam halfwidth calculation test"""
    print("\n===== Begin test_observatory_values")
    obs = Observatory(telescope_id=0)
    print(obs.get_string())
    assert obs.get_telescope_name() == 'Fake', error_msg("Incorrect telescope name")
    assert obs.get_xyz_coords() == [0.0, 0.0, 0.0], error_msg("Incorrect XYZ coords")

    gbt = Observatory(telescope_id=6)
    beam_halfwidth = gbt.calc_beam_halfwidth(100)
    assert (beam_halfwidth - 3710.19799582) < .0000001, \
        error_msg("Incorrect beam haflwidth calculation")
    print("===== End test_observatory_values")


def test_observatory_procs():
    r""" Try the member functions with Parkes """
    print("\n===== Begin test_observatory_procs")
    obs = Observatory(telescope_id=4)
    print(obs.get_string())
    print(obs.get_telescope_name())
    print(obs.get_telescope_name_short())
    print(obs.get_telescope_id())
    print(obs.get_xyz_coords())
    print(obs.get_dish_diameter())
    print(obs.get_string())
    print("===== End test_observatory_procs")


if __name__ == "__main__":
    test_observatory_construction()
    test_observatory_values()
    test_observatory_procs()



import os
import numpy as np

import blimpy as bl
from tests.data import voyager_raw, voyager_block1

def test_guppi():

    gr = bl.guppi.GuppiRaw(voyager_raw)
    header, data_idx = gr.read_header()
    nchans = header['OBSNCHAN']
    assert (nchans == 64), 'test_guppi: OBSNCHAN should be 64 but observed to be {}'.format(nchans)

    h1, data_block_x1, data_block_y1 = gr.read_next_data_block_int8()
    h2, data_block_x2, data_block_y2 = gr.read_next_data_block_int8()

    assert not np.array_equal(data_block_x1, data_block_x2) and not np.array_equal(data_block_y1, data_block_y2) \
        , "test_guppi: Data read from two blocks should not be equal"

    gr = bl.guppi.GuppiRaw(voyager_raw)
    h1, data_block_1 = gr.read_next_data_block()

    data_block_reference_1 = np.load(voyager_block1)

    assert np.array_equal(data_block_1[:, :1000, :], data_block_reference_1) \
        , "test_guppi: Data read should be consistent with previous versions"

    data_block_casted_1 = np.append(data_block_x1, data_block_y1, axis=2).astype('float32').view('complex64')

    assert np.array_equal(data_block_1, data_block_casted_1) \
        , "test_guppi: Reading as int8 then casting should be equal to reading directly as complex64"

# We have to keep instantiating objects because
# the plotting routines read data in a manner
# destructive to the object.    

def test_spectrum():
    gr = bl.guppi.GuppiRaw(voyager_raw)
    gr.plot_spectrum(flag_show=False)

def test_histogram():
    gr = bl.guppi.GuppiRaw(voyager_raw)
    gr.plot_histogram(flag_show=False)

def test_statistics():
    gr = bl.guppi.GuppiRaw(voyager_raw)
    gr.print_stats()

def test_fil_header():
    gr = bl.guppi.GuppiRaw(voyager_raw)
    header = gr.generate_filterbank_header()
    print("Generated header:\n", header)

def test_rawhdr():
    from blimpy.rawhdr import cmd_tool
    args = [voyager_raw]
    cmd_tool(args)

def test_get_obsnchan():
    nchans = bl.rawhdr.get_obsnchan(voyager_raw)
    assert (nchans == 64), 'test_get_obsnchan: OBSNCHAN should be 64 but observed to be {}'.format(nchans)

if __name__ == "__main__":
    test_guppi()
    test_spectrum()
    test_histogram()
    test_statistics()
    test_fil_header()
    test_rawhdr()
    test_get_obsnchan()




from blimpy import dice
import pytest

from tests.data import voyager_h5
from tests.data import voyager_fil

# This section makes sure that the tool exits correctly
# when certain issues are encountered.

def test_no_args():
    """
    Make sure the tool closes if no arguments are passed in
    """
    with pytest.raises(SystemExit):
        dice.cmd_tool()

def test_no_in_file():
    """
    This test does not actually increase coverage,
    which causes me to doubt the effectiveness
    of this approach to testing for errors...
    """
    with pytest.raises(SystemExit):
        args = ['-b 8419.24', '-e 8419.35', '-x', 'h5', '-otest_dice.h5']
        dice.cmd_tool()
    with pytest.raises(SystemExit):
        args = ['-f%s' % voyager_h5, '-b 8419.24', '-e 8419.35', '-x']
        dice.cmd_tool()

# This section makes sure that the tool can handle various
# file formats without exception.

def test_h5():
    args = ['-f%s' % voyager_h5, '-b 8419.24', '-e 8419.35', '-x', 'h5', '-otest_dice.h5']
    dice.cmd_tool(args)

def test_h5_no_out_file():
    args = ['-f%s' % voyager_h5, '-b 8419.24', '-e 8419.35', '-x', 'h5']
    dice.cmd_tool(args)

def test_fil():
    args = ['-f%s' % voyager_fil, '-b 8419.24', '-e 8419.35', '-x', 'fil', '-otest_dice.fil']
    dice.cmd_tool(args)

def test_fil_no_out_file():
    args = ['-f%s' % voyager_fil, '-b 8419.24', '-e 8419.35', '-x', 'fil']
    dice.cmd_tool(args)

if __name__ == "__main__":
    test_dice()



"""
Very small module with one test: test_write_to_fil()
"""
import os
import blimpy as bl

from tests.data import voyager_h5

OUTDIR = os.path.dirname(voyager_h5) + "/"

def test_write_to_fil():
    """ Load Voyager dataset and test plotting """

    a = bl.Waterfall(voyager_h5)
    a.write_to_fil(OUTDIR + 'test_out.fil')

if __name__ == "__main__":
    test_write_to_fil()



"""
Procedures for writing to a Filterbank File
"""
import time
import numpy as np
from .sigproc import generate_sigproc_header


def write_to_fil(wf, filename_out, *args, **kwargs):
    """ Write data to .fil file.
        It check the file size then decides how to write the file.

    Args:
        filename_out (str): Name of output file
    """

    # For timing how long it takes to write a file.
    t0 = time.time()

    # Update header
    wf._update_header()

    if wf.container.isheavy():
        __write_to_fil_heavy(wf, filename_out)
    else:
        __write_to_fil_light(wf, filename_out)

    t1 = time.time()
    wf.logger.info('Conversion time: %2.2fsec' % (t1 - t0))


def __write_to_fil_heavy(wf, filename_out):
    """ Write data to .fil file.

    Args:
        filename_out (str): Name of output file
    """

    # Note that a chunk is not a blob!!
    chunk_dim = wf._get_chunk_dimensions()
    blob_dim = wf._get_blob_dimensions(chunk_dim)
    n_blobs = wf.container.calc_n_blobs(blob_dim)

    # Calculate number of bytes per data element
    n_bytes = wf.header['nbits'] / 8

    wf.logger.info('Using %i n_blobs to write the data.' % n_blobs)
    with open(filename_out, "wb") as fileh:

        # Write header of .fil file
        fileh.write(generate_sigproc_header(wf))

        # For each blob
        for ii in range(0, n_blobs):

            wf.logger.info('Reading %i of %i' % (ii + 1, n_blobs))
            bob = wf.container.read_blob(blob_dim, n_blob=ii)

            # Write data of .fil file.
            if n_bytes == 4:
                np.float32(bob.ravel()).tofile(fileh)
            elif n_bytes == 2:
                np.int16(bob.ravel()).tofile(fileh)
            elif n_bytes == 1:
                np.int8(bob.ravel()).tofile(fileh)


def __write_to_fil_light(wf, filename_out):
    """ Write data to .fil file.

    Args:
        filename_out (str): Name of output file
    """

    n_bytes = wf.header['nbits'] / 8
    with open(filename_out, "wb") as fileh:
        fileh.write(generate_sigproc_header(wf))
        if n_bytes == 4:
            np.float32(wf.data.ravel()).tofile(fileh)
        elif n_bytes == 2:
            np.int16(wf.data.ravel()).tofile(fileh)
        elif n_bytes == 1:
            np.int8(wf.data.ravel()).tofile(fileh)



from .config import *
from ..utils import rebin, db


def plot_spectrum(wf, t=0, f_start=None, f_stop=None, logged=False, if_id=0, c=None, **kwargs):
    """ Plot frequency spectrum of a given file

    Args:
        t (int): integration number to plot (0 -> len(data))
        logged (bool): Plot in linear (False) or dB units (True)
        if_id (int): IF identification (if multiple IF signals in file)
        c: color for line
        kwargs: keyword args to be passed to matplotlib plot()
    """
    if wf.header['nbits'] <= 2:
        logged = False
        t = 'all'
    ax = plt.gca()

    plot_f, plot_data = wf.grab_data(f_start, f_stop, if_id)

    # Using accending frequency for all plots.
    if wf.header['foff'] < 0:
        plot_data = plot_data[..., ::-1]  # Reverse data
        plot_f = plot_f[::-1]

    if isinstance(t, int):
        print("extracting integration %i..." % t)
        plot_data = plot_data[t]
    elif t == 'all':
        print("averaging along time axis...")
        # Since the data has been squeezed, the axis for time goes away if only one bin, causing a bug with axis=1
        if len(plot_data.shape) > 1:
            plot_data = plot_data.mean(axis=0)
        else:
            plot_data = plot_data.mean()
    else:
        raise RuntimeError("Unknown integration %s" % t)

    # Rebin to max number of points
    dec_fac_x = 1
    if plot_data.shape[0] > MAX_PLT_POINTS:
        dec_fac_x = int(plot_data.shape[0] / MAX_PLT_POINTS)

    plot_data = rebin(plot_data, dec_fac_x, 1)
    plot_f = rebin(plot_f, dec_fac_x, 1)

    if not c:
        kwargs['c'] = '#333333'

    if logged:
        plt.plot(plot_f, db(plot_data), label='Stokes I', **kwargs)
        plt.ylabel("Power [dB]")
    else:

        plt.plot(plot_f, plot_data, label='Stokes I', **kwargs)
        plt.ylabel("Power [counts]")
    plt.xlabel("Frequency [MHz]")
    plt.legend()

    try:
        plt.title(wf.header['source_name'])
    except KeyError:
        plt.title(wf.filename)

    plt.xlim(plot_f[0], plot_f[-1])



#!/usr/bin/env python3

r"""
Make waterfall plots of a file set, view from top to bottom.
"""

import os
import sys
from os.path import dirname, abspath, isdir
import gc
from argparse import ArgumentParser
import logging
logger_name = "stax"
logger = logging.getLogger(logger_name)
logger.setLevel(logging.INFO)

# Plotting packages import
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("agg")

# Math/Science package imports
import numpy as np

# Blimpy imports
import blimpy as bl
from blimpy.utils import rebin

# Plotting constants
fontsize=16
font = {"family" : "DejaVu Sans",
"size" : fontsize}
MAX_IMSHOW_POINTS = (4096, 1268)


def plot_waterfall(wf, f_start=None, f_stop=None, **kwargs):
    r"""
    Plot waterfall of data in a .fil or .h5 file.

    Parameters
    ----------
    wf : blimpy.Waterfall object
        Waterfall object of an H5 or Filterbank file containing the dynamic spectrum data.
    f_start : float
        Start frequency, in MHz.
    f_stop : float
        Stop frequency, in MHz.
    kwargs : dict
        Keyword args to be passed to matplotlib imshow().

    Notes
    -----
    Plot a single-panel waterfall plot (frequency vs. time vs. intensity)
    for one of the files in the set of interest, at the
    frequency of the expected event.
    """

    # Extract source name.
    source_name = wf.header["source_name"]

    # prepare font
    matplotlib.rc("font", **font)

    # Load in the data from fil
    plot_f, plot_data = wf.grab_data(f_start=f_start, f_stop=f_stop)

    # Make sure waterfall plot is under 4k*4k
    dec_fac_x, dec_fac_y = 1, 1

    # rebinning data to plot correctly with fewer points
    try:
        if plot_data.shape[0] > MAX_IMSHOW_POINTS[0]:
            dec_fac_x = plot_data.shape[0] / MAX_IMSHOW_POINTS[0]
        if plot_data.shape[1] > MAX_IMSHOW_POINTS[1]:
            dec_fac_y =  int(np.ceil(plot_data.shape[1] /  MAX_IMSHOW_POINTS[1]))
        plot_data = rebin(plot_data, dec_fac_x, dec_fac_y)
    except Exception as ex:
        print("\n*** Oops, grab_data returned plot_data.shape={}, plot_f.shape={}"
              .format(plot_data.shape, plot_f.shape))
        print("Waterfall info for {}:".format(wf.filename))
        wf.info()
        raise ValueError("*** Something is wrong with the grab_data output!") from ex

    # Rolled back PR #82

    # determine extent of the plotting panel for imshow
    extent=(plot_f[0], plot_f[-1], (wf.timestamps[-1]-wf.timestamps[0])*24.*60.*60, 0.0)

    # plot and scale intensity (log vs. linear)
    kwargs["cmap"] = kwargs.get("cmap", "viridis")
    plot_data = 10.0 * np.log10(plot_data)

    # get normalization parameters
    vmin = plot_data.min()
    vmax = plot_data.max()
    normalized_plot_data = (plot_data - vmin) / (vmax - vmin)

    # display the waterfall plot
    this_plot = plt.imshow(normalized_plot_data,
        aspect="auto",
        rasterized=True,
        interpolation="nearest",
        extent=extent,
        **kwargs
    )

    # add plot labels
    plt.xlabel("Frequency [Hz]",fontdict=font)
    plt.ylabel("Time [s]",fontdict=font)

    # add source name
    ax = plt.gca()
    plt.text(0.03, 0.8, source_name, transform=ax.transAxes, bbox=dict(facecolor="white"))

    return this_plot


def sort2(x, y):
    r""" Return lowest value, highest value"""
    if y < x:
        return y, x
    return x, y


def make_waterfall_plots(file_list, plot_dir, f_start=None, f_stop=None, **kwargs):
    r"""
    Make waterfall plots of a file set, view from top to bottom.

    Parameters
    ----------
    file_list : str
        List of filterbank files to plot in a stacked mode.
    f_start : float
        Start frequency, in MHz.
    f_stop : float
        Stop frequency, in MHz.
    source_name_list : list
        List of source names in the set, in order.
    kwargs : dict
        Keyword args to be passed to matplotlib imshow().
    """

    # prepare for plotting
    matplotlib.rc("font", **font)

    # set up the sub-plots
    n_plots = len(file_list)
    fig = plt.subplots(n_plots, sharex=True, sharey=True,figsize=(10, 2*n_plots))

    # get directory path for storing PNG files
    if plot_dir is None:
        dirpath = dirname(abspath(file_list[0])) + "/"
    else:
        if not isdir(plot_dir):
            os.mkdir(plot_dir)
        dirpath = plot_dir


    # read in data for the first panel
    max_load = bl.calcload.calc_max_load(file_list[0])
    #print("plot_event make_waterfall_plots: max_load={} is required for {}".format(max_load, file_list[0]))
    wf1 = bl.Waterfall(file_list[0], max_load=max_load)
    t0 = wf1.header["tstart"]
    source_name = wf1.header["source_name"]

    # Fix frequency boundaries if required.
    freqs = wf1.container.populate_freqs()
    if f_start is None:
        f_start = freqs[0]
    if f_stop is None:
        f_stop = freqs[-1]
    the_lowest, the_highest = sort2(f_start, f_stop)

    # rebin data to plot correctly with fewer points
    plot_f1, plot_data1 = wf1.grab_data(f_start=the_lowest, f_stop=the_highest)
    dec_fac_x, dec_fac_y = 1, 1
    if plot_data1.shape[0] > MAX_IMSHOW_POINTS[0]:
        dec_fac_x = plot_data1.shape[0] / MAX_IMSHOW_POINTS[0]
    if plot_data1.shape[1] > MAX_IMSHOW_POINTS[1]:
        dec_fac_y =  int(np.ceil(plot_data1.shape[1] /  MAX_IMSHOW_POINTS[1]))
    plot_data1 = rebin(plot_data1, dec_fac_x, dec_fac_y)

    # Compute the midpoint.
    the_midpoint = np.abs(the_lowest + the_highest) / 2.

    # Protect RAM utilisation.
    del wf1, freqs, plot_f1, plot_data1
    gc.collect()

    # Fill in each subplot for the full plot
    subplots = []
    for ii, filename in enumerate(file_list):
        logger.debug("make_waterfall_plots: file {} in list: {}".format(ii, filename))
        # identify panel
        subplot = plt.subplot(n_plots, 1, ii + 1)
        subplots.append(subplot)

        # read in data
        max_load = bl.calcload.calc_max_load(filename)
        wf = bl.Waterfall(filename, max_load=max_load)

        # Validate frequency range.
        freqs = wf.container.populate_freqs()
        ii_lowest, ii_highest = sort2(freqs[0], freqs[-1])
        logger.info("Processing: {}, freq lowest={}, highest={}".format(filename, ii_lowest, ii_highest))
        if the_lowest < ii_lowest or the_highest > ii_highest:
            logger.warning("Frequency range not compatible!  Ignoring this file.")
            # Protect RAM utilisation.
            del wf, freqs
            gc.collect()
            continue # Skip this file.

        # make plot with plot_waterfall
        source_name = wf.header["source_name"]
        this_plot = plot_waterfall(wf,
                                   f_start=the_lowest,
                                   f_stop=the_highest,
                                   **kwargs)

        # Title the full plot if processing the first file.
        if ii == 0:
            plot_title = "%s \n MJD:%5.5f (first file)" % (source_name, t0)
            plt.title(plot_title)

        # Format full plot.
        if ii < len(file_list)-1:
            plt.xticks(np.linspace(the_lowest, the_highest, num=4), ["","","",""])

        # Protect RAM utilisation.
        del wf, freqs
        gc.collect()

    # More overall plot formatting, axis labelling.
    factor = 1e6
    units = "Hz"
    xloc = np.linspace(the_lowest, the_highest, 5)
    xticks = [round(loc_freq) for loc_freq in (xloc - the_midpoint) * factor]
    if np.max(xticks) > 1000:
        xticks = [xt / 1000 for xt in xticks]
        units = "kHz"
    plt.xticks(xloc, xticks)
    plt.xlabel("Relative Frequency [%s] from %f MHz" % (units, the_midpoint), fontdict=font)

    # Add colorbar.
    cax = fig[0].add_axes([0.94, 0.11, 0.03, 0.77])
    fig[0].colorbar(this_plot, cax=cax, label="Normalized Power (Arbitrary Units)")

    # Adjust plots
    plt.subplots_adjust(hspace=0,wspace=0)

    # Save the figures.
    path_png = dirpath + source_name + "_fstart_{:0.6f}".format(f_start) \
                + "_fstop_{:0.6f}".format(f_stop) + ".png"
    plt.savefig(path_png, bbox_inches="tight")
    logger.info("Saved plot: {}".format(path_png))

    # show figure before closing if this is an interactive context
    mplbe = matplotlib.get_backend()
    logger.debug("make_waterfall_plots: backend = {}".format(mplbe))
    if mplbe != "agg":
        plt.show()

    # close all figure windows
    plt.close("all")


def cmd_line(args=None):
    r"""Coomand line parser"""
    parser = ArgumentParser(description='Make waterfall plots from a set of files, viewed from top to bottom.')
    parser.add_argument('file_list', type=str, nargs='+', help='List of files to plot')
    parser.add_argument('-f_start', type=float, default=None, help='Start frequency.  Default: None.')
    parser.add_argument('-f_stop', type=float, default=None, help='Stop frequency.  Default: None.')
    parser.add_argument('-plot_dir', type=str, default=".", help='Directory to receive the plot (.png).  Default: current directory.')
    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)
    if args.file_list == []:
        os.system("stax -h")
        sys.exit(0)

    # Make the plots.
    args.plot_dir += "/"
    args.file_list = sorted(args.file_list)
    logger.info("Begin")
    make_waterfall_plots(args.file_list, args.plot_dir, args.f_start, args.f_stop)
    logger.info("End")


if __name__ == "__main__":
    cmd_line()



r'''
Read the specified raw file.
Print the entire header.
Highlight the number of observation channels.
'''

import sys
import blimpy
from argparse import ArgumentParser


def get_obsnchan(filepath):
    gr = blimpy.GuppiRaw(filepath)
    header, data_idx = gr.read_header()
    return header['OBSNCHAN']


def cmd_tool(args=None):
    p = ArgumentParser(description='Calculate the Waterfall max_load value needed to load the data array for a given file.')
    p.add_argument('filepath', type=str, help='Name of raw guppi file path to access')   
    if args is None:
        args = p.parse_args()
    else:
        args = p.parse_args(args)

    print('Getting the header from {}.\n'.format(args.filepath))
    with open(args.filepath, 'rb') as fh:
        while True:
            buffer = fh.read(80).decode("utf-8").strip()

            print('\t', buffer)
            if buffer[0:3] == "END":
                break
    print('\nThe number of observation channels:', get_obsnchan(args.filepath))

if __name__ == "__main__":
    cmd_tool()



# the lack of some implementations causes a maximum of
# 18 guaranteed misses in codecov

import blimpy as bl
import pytest
from tests.data import voyager_fil, voyager_h5, here

def test_read_fns():
    """ These read functions are currently not implemented. """
    """ L: if that is true, why do the first two lines
        NOT raise NotImplementedError's?"""
    a = bl.Waterfall(voyager_fil)
    b = bl.Waterfall(voyager_h5)
    with pytest.raises(NotImplementedError):
        a.container.read_all()
        a.container.read_row(0)
        a.container.read_rows(0, 2)

        b.container.read_all()
        b.container.read_row(0)
        b.container.read_rows(0, 2)

def test_file_wrapper_open_h5_file():
    from blimpy.io.file_wrapper import open_file
    h5r = open_file(voyager_h5)

    # check all if branches
    h5r.populate_timestamps(update_header=False)
    h5r.populate_timestamps(update_header=True)
  
    with pytest.raises(ValueError):
        h5r.read_blob((300, 300, 300, 300), -1)        

    # testing to make sure all internal functions are already used
    h5r._setup_selection_range(
        f_start=3, f_stop=4, t_stop=3, t_start=4
    )
    h5r._setup_selection_range()    

    h5r._init_empty_selection()
    
    # perform reading again and we make sure that
    # the user-facing functions survive

    h5r.populate_freqs()
    
    # just arbitrary values
    h5r.calc_n_blobs(300)
    h5r.read_blob((300, 300, 300, 300))

    h5r.calc_n_coarse_chan()
    h5r.calc_n_coarse_chan(1)
    # we have at least 2 if branches that we are not exploring

def test_file_wrapper_open_fil_file():
    from blimpy.io.file_wrapper import open_file
    filr = open_file(voyager_fil)
    
    filr.calc_n_blobs(300)
    filr.read_blob((300, 300, 300, 300))

def test_invalid_files():
    from blimpy.io.file_wrapper import open_file
    
    # script should error-out if file does not exist
    with pytest.raises(IOError):
        open_file(here + 'file_does_not_exist.h5')

    # script should error-out if file is not a
    # valid blimpy data file
    with pytest.raises(NotImplementedError):
        open_file(here + '/run_tests.sh')

if __name__ == "__main__":
    test_read_fns()
    test_file_wrapper_open_file()



from os import path

here = path.dirname(path.abspath(__file__))

voyager_fil = path.join(here, 'test_data/Voyager1.single_coarse.fine_res.fil')
voyager_h5 = path.join(here, 'test_data/Voyager1.single_coarse.fine_res.h5')
voyager_raw = path.join(here, 'test_data/blc3_2bit_guppi_57396_VOYAGER1_0006.0013.raw')
voyager_block1 = path.join(here, 'test_data/Voyager1_block1.npy')
test_ifs_fil = path.join(here, 'test_data/test_ifs.fil')



"""
# utils.py
useful helper functions for common data manipulation tasks
"""
import numpy as np


def db(x, offset=0):
    """ Convert linear to dB """
    return 10 * np.log10(x + offset)


def lin(x):
    """ Convert dB to linear """
    return 10.0 ** (x / 10.0)


def closest(xarr, val):
    """ Return the index of the closest in xarr to value val """
    idx_closest = np.argmin(np.abs(np.array(xarr) - val))
    return idx_closest


def rebin(d, n_x=None, n_y=None, n_z=None):
    """ Rebin data by averaging bins together

    Args:
    d (np.array): data
    n_x (int): number of bins in x dir to rebin into one
    n_y (int): number of bins in y dir to rebin into one

    Returns:
    d: rebinned data with shape (n_x, n_y)
    """
    n_x = 1 if n_x is None else n_x
    n_y = 1 if n_y is None else n_y
    n_z = 1 if n_z is None else n_z

    if d.ndim == 3:
        d = d[:int(d.shape[0] // n_x) * n_x, :int(d.shape[1] // n_y) * n_y, :int(d.shape[2] // n_z) * n_z]
        d = d.reshape((d.shape[0] // n_x, n_x, d.shape[1] // n_y, n_y, d.shape[2] // n_z, n_z))
        d = d.mean(axis=5)
        d = d.mean(axis=3)
        d = d.mean(axis=1)
    elif d.ndim == 2:
        d = d[:int(d.shape[0] // n_x) * n_x, :int(d.shape[1] // n_y) * n_y]
        d = d.reshape((d.shape[0] // n_x, n_x, d.shape[1] // n_y, n_y))
        d = d.mean(axis=3)
        d = d.mean(axis=1)
    elif d.ndim == 1:
        d = d[:int(d.shape[0] // n_x) * n_x]
        d = d.reshape((d.shape[0] // n_x, n_x))
        d = d.mean(axis=1)
    else:
        raise RuntimeError("Only NDIM <= 3 supported")
    return d


def unpack(data, nbit):
    """upgrade data from nbits to 8bits

    Notes: Pretty sure this function is a little broken!
    """
    if nbit > 8:
        raise ValueError("unpack: nbit must be <= 8")
    if 8 % nbit != 0:
        raise ValueError("unpack: nbit must divide into 8")
    if data.dtype not in (np.uint8, np.int8):
        raise TypeError("unpack: dtype must be 8-bit")
    if nbit == 8:
        return data
    elif nbit == 4:
        data = unpack_4to8(data)
        return data
    elif nbit == 2:
        data = unpack_2to8(data)
        return data
    elif nbit == 1:
        data = unpack_1to8(data)
        return data


def unpack_1to8(data):
    """ Promote 1-bit unisgned data into 8-bit unsigned data.

    Args:
        data: Numpy array with dtype == uint8
    """
    return np.unpackbits(data)


def unpack_2to8(data):
    """ Promote 2-bit unisgned data into 8-bit unsigned data.

    Args:
        data: Numpy array with dtype == uint8

    Notes:
        DATA MUST BE LOADED as np.array() with dtype='uint8'.

        This works with some clever shifting and AND / OR operations.
        Data is LOADED as 8-bit, then promoted to 32-bits:
        /ABCD EFGH/ (8 bits of data)
        /0000 0000/0000 0000/0000 0000/ABCD EFGH/ (8 bits of data as a 32-bit word)

        Once promoted, we can do some shifting, AND and OR operations:
        /0000 0000/0000 ABCD/EFGH 0000/0000 0000/ (shifted << 12)
        /0000 0000/0000 ABCD/EFGH 0000/ABCD EFGH/ (bitwise OR of previous two lines)
        /0000 0000/0000 ABCD/0000 0000/0000 EFGH/ (bitwise AND with mask 0xF000F)
        /0000 00AB/CD00 0000/0000 00EF/GH00 0000/ (prev. line shifted << 6)
        /0000 00AB/CD00 ABCD/0000 00EF/GH00 EFGH/ (bitwise OR of previous two lines)
        /0000 00AB/0000 00CD/0000 00EF/0000 00GH/ (bitwise AND with 0x3030303)

        Then we change the view of the data to interpret it as 4x8 bit:
        [000000AB, 000000CD, 000000EF, 000000GH]  (change view from 32-bit to 4x8-bit)

        The converted bits are then mapped to values in the range [-40, 40] according to a lookup chart.
        The mapping is based on specifications in the breakthough docs:
        https://github.com/UCBerkeleySETI/breakthrough/blob/master/doc/RAW-File-Format.md

    """
    two_eight_lookup = {0: 40,
                        1: 12,
                        2: -12,
                        3: -40}

    tmp = data.astype(np.uint32)
    tmp = (tmp | (tmp << 12)) & 0xF000F
    tmp = (tmp | (tmp << 6)) & 0x3030303
    tmp = tmp.byteswap()
    tmp = tmp.view('uint8')
    mapped = np.array(tmp, dtype=np.int8)
    for k, v in two_eight_lookup.items():
        mapped[tmp == k] = v
    return mapped


def unpack_4to8(data):
    """ Promote 2-bit unisgned data into 8-bit unsigned data.

    Args:
        data: Numpy array with dtype == uint8

    Notes:
        # The process is this:
        # ABCDEFGH [Bits of one 4+4-bit value]
        # 00000000ABCDEFGH [astype(uint16)]
        # 0000ABCDEFGH0000 [<< 4]
        # 0000ABCDXXXXEFGH [bitwise 'or' of previous two lines]
        # 0000111100001111 [0x0F0F]
        # 0000ABCD0000EFGH [bitwise 'and' of previous two lines]
        # ABCD0000EFGH0000 [<< 4]
        # which effectively pads the two 4-bit values with zeros on the right
        # Note: This technique assumes LSB-first ordering
    """

    tmpdata = data.astype(np.int16)  # np.empty(upshape, dtype=np.int16)
    tmpdata = (tmpdata | (tmpdata << 4)) & 0x0F0F
    # tmpdata = tmpdata << 4 # Shift into high bits to avoid needing to sign extend
    updata = tmpdata.byteswap()
    return updata.view(data.dtype)



r"""
    From an input HDF file or Filterbank file, perform channel averaging,
    producing a new HDF5 file of Filterbank file.
"""

import os
from argparse import ArgumentParser
from blimpy.waterfall import Waterfall


def bl_scrunch(in_path, out_dir='./', new_filename='', max_load=None, f_scrunch=None):
    r""" Frequency scrunch (lower resolution by averaging)

    Args:
        in_path : str
            Path of input file to open.
        out_dir : str
            Output directory.
        new_filename : str
            Output file name.
        max_load : int
            Waterfall object instantiation max_load parameter value.
        f_scrunch : int
            Number of frequency channels to average together at one time.
    """

    print("bl_scrunch: Input path: {}".format(in_path))
    in_ext = os.path.splitext(in_path)[1]
    if in_ext not in ('.fil', '.h5'):
        raise ValueError('Oops, input file extension must be .fil or .h5; saw: {} !'.format(in_ext))
    print("bl_scrunch: Averaging {} frequency channels at a time.".format(f_scrunch))

    wf = Waterfall(in_path, max_load=max_load)
    if new_filename == '':
        if in_ext == '.h5':
            out_path = out_dir + '/' + in_path.replace('.h5', '.scrunched.h5').split('/')[-1]
        else: # .fil
            out_path = out_dir + '/' + in_path.replace('.fil', '.scrunched.h5').split('/')[-1]
    else:
        out_path = out_dir + new_filename

    print("bl_scrunch: Output path: {}".format(out_path))
    wf.write_to_hdf5(out_path, f_scrunch=f_scrunch)
    print("bl_scrunch: End")


def cmd_tool(args=None):
    r"""  Command line utility for scrunching an input HDF5 file or Filterbank file.
    """

    p = ArgumentParser(description='Command line utility for converting HDF5 (.h5) to Sigproc filterbank (.fil) format \n >>h52fil <FULL_PATH_TO_FIL_FILE> [options]')
    p.add_argument('filepath', type=str, help='Name of file path to open (.h5 or .fil).')
    p.add_argument('-f', '--fscrunch', dest='f_scrunch', type=int,
                 help='Number of frequency channels to average (scrunch) together.')
    p.add_argument('-o', '--out_dir', dest='out_dir', type=str, default='./',
                 help='Location for output files. Default: current directory.')
    p.add_argument('-n', '--new_filename', dest='new_filename', type=str, default='',
                 help='New filename. Default: replaces the file extension with .scrunched.fil or .scrunched .h5.')
    p.add_argument('-l', '--max_load', action='store', default=None, dest='max_load', type=float,
                 help='Maximum data limit to load. Default: 1.0 GB.')
    p.add_argument('-d', '--delete_input', dest='delete_input', action='store_true', default=False,
                 help='This option deletes the input file after conversion.')

    if args is None:
        args = p.parse_args()
    else:
        args = p.parse_args(args)

    bl_scrunch(args.filepath, out_dir=args.out_dir, new_filename=args.new_filename,
               max_load=args.max_load, f_scrunch=args.f_scrunch)

    if args.delete_input:
        print("'Deleting input file: %s"%(args.filepath))
        os.remove(args.filepath)


if __name__ == "__main__":

    cmd_tool()



r""" De-doppler signal processing functions """


import numpy as np


def dedoppler_1(wf, drift_rate):
    """
    Simple de-doppler code for a Filterbank or HDF5 file.
    Parameters:
    ----------
    wf : object
        Blimpy Waterfall object, previously instantiated with a loaded data matrix.
    drift_rate : float
        Signal drift rate over time [Hz/s]
    """

    # Get the time sampling interval in seconda.
    tsamp = wf.header['tsamp']

    # Get the fine channel bandwidth in Hz.
    chan_bw = wf.header['foff'] * 1e6

    # Compute the number of numpy rolls to perform.
    n_roll = (drift_rate * tsamp) / chan_bw

    # For each time-row,
    #      roll all of the data power values in each fine channel frequency column
    #      given by -(n_roll * row number).
    for ii in range(wf.data.shape[0]):
        wf.data[ii][0][:] = np.roll(wf.data[ii][0][:], -int(n_roll * ii))



from astropy.io import ascii
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit

data = ascii.read('calibrators.txt')


def func_powerlaw(x,alph,c):
    """
    Power-law function
    """
    return c*x**alph

def get_freqs(MHz=False):
    """
    Return frequency array in calibrators.txt
    If MHz is true, then units of values are in MHz;
    Else the units are in GHz (default).
    """
    if MHz:
        return np.array(data['Freq']) * 1000
    return np.array(data['Freq'])

def cal_fluxes(source):
    """
    Return spectrum of a particular source (e.g. '3C295') in calibrators.txt
    """
    return np.array(data[source])

def errors():
    """
    Get errors in the data matrix.
    """
    return np.array(data['Err'])

def source_power_law_fit(source,minfreq,maxfreq):
    """
    Calculates optimized power-law parameters. Considers data for a particular source
    from calibrators.txt between minfreq and maxfreq (in MHz)
    """
    freqs = get_freqs(MHz=True) #In MHz
    fluxes = cal_fluxes(source)
    freqs_cut = freqs[np.where(np.logical_and(freqs>=minfreq, freqs<=maxfreq))]
    fluxes_cut = fluxes[np.where(np.logical_and(freqs>=minfreq, freqs<=maxfreq))]
    popt, dummy = curve_fit(func_powerlaw,freqs_cut,fluxes_cut)
    return popt

def plot_flux_comp(source,name=None,custom_minfreq=None,custom_maxfreq=None):
    """
    Plots the result of source_power_law_fit() along with the data from calibrators.txt
    for the source in question. Use 'name' to save the resulting plot.
    """
    freqs = get_freqs(MHz=True) #In MHz
    fluxes = cal_fluxes(source)
    errs = errors()
    if custom_minfreq is not None:
        minfreq = custom_minfreq
    else: minfreq = freqs.min()

    if custom_maxfreq is not None:
        maxfreq = custom_maxfreq
    else: maxfreq = freqs.max()

    alph,c = source_power_law_fit(source,minfreq,maxfreq)
    print(alph)

    freqvec = np.linspace(freqs.min(),freqs.max(),5000)
    model = func_powerlaw(freqvec,alph,c)

    plt.plot(freqvec,model,'r-',label='Power-law model')
    plt.errorbar(freqs,fluxes,yerr=fluxes*errs/100.0,marker='o',mfc='blue',markersize=2.5,capsize=2,capthick=0.5,fmt='go',label='Actual values')

    plt.xlabel('Frequency (MHz)')
    plt.ylabel('Flux (Jy)')
    plt.yscale('log')
    plt.title('Predicted and Actual Spectrum of '+source)
    plt.legend()
    if name is not None:
        plt.savefig(name,dpi=250)
    plt.show()



r"""
test_bl_scrunch
"""

import os
import pytest

import blimpy as bl
from tests.data import voyager_h5

OUTDIR = os.path.dirname(voyager_h5) + "/"
print("test_bl_scrunch: output directory:", OUTDIR)


def test_scrunch():
    r"""
    Tests the conversion of fil files into h5 in both light and heavy modes.
    But apparently it does not test the accuracy of the conversion.
    """
    print("\n===== test_scrunch BEGIN")
    # Creating test file.
    bl.bl_scrunch.bl_scrunch(voyager_h5, new_filename='test.scrunched.h5', f_scrunch=8)

    # Deleting test file
    os.remove('test.scrunched.h5')
    print("\n===== test_scrunch END")


def test_nameless():
    r"""
    The script should trigger a system
    exit if the user fails to provide a file name.
    """

    print("\n===== test_nameless BEGIN")
    bl.bl_scrunch.bl_scrunch(voyager_h5, out_dir=OUTDIR, f_scrunch=8)

    with pytest.raises(SystemExit):
        args = ["-o", OUTDIR, "-n", "test.scrunched.h5", "-f", "8", "-l", "0.1"]
        bl.bl_scrunch.cmd_tool(args)
    print("\n===== test_nameless END")


def test_cmd():
    r""" Pass in some example sets of arguments """
    # this is a malformed arg set with one giant string in the first entry
    print("\n===== test_cmd BEGIN")
    args = [voyager_h5, "-o", OUTDIR, "-n", "test.scrunched.h5", "-f", "8", "-l", "0.1"]
    bl.bl_scrunch.cmd_tool(args)
    args = [voyager_h5, "-f", "8", "-l", "0.1", "--out_dir", OUTDIR]
    bl.bl_scrunch.cmd_tool(args)
    args = [voyager_h5, "--mickey_mouse", "-f", "8", "-l", "0.1", "-o", OUTDIR]
    with pytest.raises(SystemExit):
        bl.bl_scrunch.cmd_tool(args)
    print("\n===== test_cmd END")


if __name__ == "__main__":
    test_scrunch()
    test_nameless()
    test_cmd()



from .plot_waterfall import plot_waterfall
from .plot_spectrum import plot_spectrum
from .plot_spectrum_min_max import plot_spectrum_min_max
from .plot_kurtosis import plot_kurtosis
from .plot_time_series import plot_time_series
from .plot_all import plot_all
from . import config as plot_config
